"""Test cases for audio ingestors"""
from pathlib import Path
import pytest
import asyncio

from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import uuid


@pytest.mark.asyncio
async def test_collect_and_ingest_audio():
    collector_factory = FSCollectorFactory()
    uri = Uri("file://" + str(Path("./tests/data/audio/").resolve()))
    config = FSCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "root_path": uri.path,
            "name": "Local-config",
            "config": {},
            "uri": "file://",
        }
    )
    collector = collector_factory.resolve(uri, config)

    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("mp3")

    ingestor = await ingestor_factory.create("mp3", [])

    # Collect and ingest the PDF
    ingested_call = ingestor.ingest(collector.poll())
    counter = 0

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested.error is None
            assert ingested.file is not None
            counter += 1

        # counter is 2 because at the end of each file there is an empty IngestedTokens being yielded
        assert counter == 2

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_audio())



from enum import Enum


class ResourceConfigKey(Enum):
    ID = "id"
    MAX_WORKERS_ALLOWED = "max_workers_allowed"
    MAX_WORKERS_PER_COLLECTOR = "max_workers_per_collector"
    MAX_WORKERS_PER_ENGINE = "max_workers_per_engine"
    MAX_WORKERS_PER_QUERENT = "max_workers_per_querent"



class FunctionRegistry:
    def __init__(self):
        self._predicate_info_functions = [
            {
                "type": "function",
                "function": {
                    'name': 'predicate_info',
                    'description': 'Identify the predicate (relationship) and its type based on the context and the subject-object',
                    'parameters': {
                        'type': 'object',
                        'properties': {
                            'predicate': {
                                'type': 'string',
                                'description': 'The relationship (predicate) between the subject and the object.'
                            },
                            'predicate_type': {
                                'type': 'string',
                                'description': 'The type of the relationship (predicate type).'
                            },

                        }, "required": ["predicate", "predicate_type"],
                    }
                }
            }
        ]

        self._classifyentity_functions = [
        {
            "type": "function",
            "function": {
                'name': 'classify_entities',
                'description': 'This function identifies and classifies two provided entities as either the subject or the object within a given context. It also categorizes these entities based on their types, providing a clear semantic understanding of their roles.',
                'parameters': {
                    'type': 'object',
                    'properties': {
                        'subject': {
                            'type': 'string',
                            'description': 'Identified as the main entity in the context, this is typically the initiator or the primary focus of the action or topic being discussed.'
                        },
                        'subject_type': {
                            'type': 'string',
                            'description': 'The subject type.'
                        },
                        'object': {
                            'type': 'string',
                            'description': 'This parameter represents the entity in the context that is directly impacted by or involved in the action. It is typically the recipient or target of the main verbs action.'
                        },
                        
                        'object_type': {
                            'type': 'string',
                            'description': 'The type of the object entity.'
                        }
                    
                    }, "required": ["subject","subject_type", "object", "object_type"],
                }
            }    
        }
    ]

    def get_predicate_info_function(self):
        return self._predicate_info_functions

    def get_classifyentity_function(self):
        return self._classifyentity_functions





import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import SlackCollectorConfig
from querent.common.uri import Uri
import uuid


@pytest.fixture
def slack_config():
    return SlackCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "channel_name": "C05TA5R7D88",
            "cursor": None,
            "include_all_metadata": 0,
            "inclusive": 0,
            "latest": 0,
            "limit": 100,
            "access_token": os.getenv("SLACK_ACCESS_KEY"),
            "name": "Slack-config",
            "config": {},
            "uri": "slack://",
        }
    )


@pytest.mark.asyncio
async def test_slack_collector(slack_config):
    uri = Uri("slack://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, slack_config)
    assert collector is not None

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert not result.is_error()
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        # Changing the number to be greater than 1 because messages before 90 days gets deleted from slack
        assert counter > 0

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_slack_collector())



import asyncio
import json
from pathlib import Path
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.common.types.querent_event import EventState, EventType
from querent.common.uri import Uri
from querent.config.core.llm_config import LLM_Config
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid
from querent.core.transformers.bert_llm import BERTLLM
from querent.querent.resource_manager import ResourceManager
from querent.querent.querent import Querent

from dotenv import load_dotenv

from querent.config.collector.collector_config import (
    FSCollectorConfig,
    DriveCollectorConfig,
    SlackCollectorConfig,
)
from querent.common.uri import Uri
from querent.collectors.collector_resolver import CollectorResolver
import os

load_dotenv()


def drive_config():
    return DriveCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "drive_refresh_token": os.getenv("DRIVE_REFRESH_TOKEN"),
            "drive_token": os.getenv("DRIVE_TOKEN"),
            "drive_scopes": os.getenv("DRIVE_SCOPES"),
            "chunk_size": 1024 * 1024,
            "drive_client_id": os.getenv("DRIVE_CLIENT_ID"),
            "drive_client_secret": os.getenv("DRIVE_CLIENT_SECRET"),
            "specific_file_type": "application/pdf",
            "folder_to_crawl": "1BtLKXcYBrS16CX0R4V1X7Y4XyO9Ct7f8",
            "name": "Drive-config",
            "config": {},
            "uri": "drive://",
        }
    )


def slack_config():
    return SlackCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "channel_name": "C05TA5R7D88",
            "cursor": None,
            "include_all_metadata": 0,
            "inclusive": 0,
            "latest": 0,
            "limit": 100,
            "access_token": os.getenv("SLACK_ACCESS_KEY"),
            "name": "Slack-config",
            "config": {},
            "uri": "slack://",
        }
    )


@pytest.mark.asyncio
async def test_multiple_collectors_all_async():
    # Set up the collectors
    directories = ["./tests/data/llm/old_pdf", "./tests/data/llm/pdf"]
    collectors = [
        CollectorResolver().resolve(
            Uri("file://" + str(Path(directory).resolve())),
            FSCollectorConfig(config_source={
                    "id": str(uuid.uuid4()),
                    "root_path": directory,
                    "name": "Local-config",
                    "config": {},
                    "uri": "file://",
                }),
        )
        for directory in directories
    ]

    collectors.append(
        CollectorResolver().resolve(
            Uri("drive://"),
            drive_config(),
        )
    )

    collectors.append(
        CollectorResolver().resolve(
            Uri("slack://"),
            slack_config(),
        )
    )

    for collector in collectors:
        await collector.connect()

    # Set up the result queue
    result_queue = asyncio.Queue()

    # Create the IngestorFactoryManager
    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )

    # Start the ingest_all_async in a separate task
    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())
    resource_manager = ResourceManager()
    bert_llm_config = LLM_Config(
        ner_model_name="botryan96/GeoBERT",
        enable_filtering=True,
        filter_params={
            "score_threshold": 0.5,
            "attention_score_threshold": 0.1,
            "similarity_threshold": 0.5,
            "min_cluster_size": 5,
            "min_samples": 3,
            "cluster_persistence_threshold": 0.1,
        },
        fixed_entities=["carbon isotope", "carbon stable isotope"],
        sample_entities=["B-GeoPetro", "B-GeoPetro"],
    )
    llm_instance = BERTLLM(result_queue, bert_llm_config)

    class StateChangeCallback(EventCallbackInterface):
        async def handle_event(self, event_type: EventType, event_state: EventState):
            assert event_state.event_type == EventType.Graph
            triple = json.loads(event_state.payload)
            print("triple: {}".format(triple))
            assert isinstance(triple["subject"], str) and triple["subject"]

    llm_instance.subscribe(EventType.Graph, StateChangeCallback())
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )

    querent_task = asyncio.create_task(querent.start())
    await asyncio.gather(ingest_task, querent_task)


if __name__ == "__main__":
    asyncio.run(test_multiple_collectors_all_async())



import asyncio
import json
import pytest
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.common.types.querent_queue import QuerentQueue
from querent.common.types.querent_event import EventState, EventType
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.querent_queue import QuerentQueue
from querent.config.core.llm_config import LLM_Config
from querent.core.transformers.bert_llm import BERTLLM
from querent.querent.resource_manager import ResourceManager
from querent.querent.querent import Querent


@pytest.mark.asyncio
@pytest.mark.parametrize("input_data,ner_model_name, llm_class,expected_entities,filter_entities", [
   (["In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca. 56 Ma) was a rapid global warming event characterized by the rise of temperatures to5–9 °C (Kennett and Stott, 1991), which caused substantial environmental changes around the globe."],
"botryan96/GeoBERT", BERTLLM, ["tectonic perturbations","downstream sectors"], True)])



async def test_bertllm_ner_tokenization_and_entity_extraction(input_data, ner_model_name, llm_class, expected_entities, filter_entities):
    input_queue = QuerentQueue()
    resource_manager = ResourceManager()
    ingested_data = IngestedTokens(file="dummy_1_file.txt", data=input_data)
    await input_queue.put(ingested_data)
    ingested_data = IngestedTokens(file="dummy_1_file.txt", data=None)
    await input_queue.put(ingested_data)
    await input_queue.put(None)
    bert_llm_config = LLM_Config(
        ner_model_name=ner_model_name,
        enable_filtering=filter_entities,
        filter_params={
            'score_threshold': 0.5,
            'attention_score_threshold': 0.1,
            'similarity_threshold': 0.5,
            'min_cluster_size': 5,
            'min_samples': 3,
            'cluster_persistence_threshold':0.2
        }
    )
    llm_instance = llm_class(input_queue, bert_llm_config)
    class StateChangeCallback(EventCallbackInterface):
        def handle_event(self, event_type: EventType, event_state: EventState):
            assert event_state.event_type == EventType.Graph
            triple = json.loads(event_state.payload)
            print("triple: {}".format(triple))
            assert isinstance(triple['subject'], str) and triple['subject']
    llm_instance.subscribe(EventType.Graph, StateChangeCallback())
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )

    await querent.start()




import asyncio
from pathlib import Path
from querent.collectors.github.github_collector import GithubCollectorFactory
from querent.config.collector.collector_config import GithubConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import os
import uuid
from dotenv import load_dotenv

load_dotenv()


@pytest.mark.asyncio
async def test_collect_and_ingest_code():
    collector_factory = GithubCollectorFactory()
    config = GithubConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "github_username": os.getenv("USERNAME_GITHUB"),
            "repository": os.getenv("REPOSITORY_NAME_GITHUB"),
            "github_access_token": os.getenv("ACCESS_TOKEN_GITHUB"),
            "name": "Github-config",
            "config": {},
            "uri": "github://"
        }
    )
    uri = Uri("github://")

    collector = collector_factory.resolve(uri, config)

    # Set up the ingestor
    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("github")
    ingestor = await ingestor_factory.create("github", [])

    ingested_call = ingestor.ingest(collector.poll())
    counter = 0

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested is not None
            if ingested is not "" or ingested is not None:
                counter += 1

        # 6 extra IngestedTokens signifying end of file
        assert counter == 12

    await poll_and_print()  # Notice the use of await here


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_code())



import json
from llama_cpp import Llama
from querent.logging.logger import setup_logger

"""
    QASystem is a class designed for a question-answering system, leveraging large language models (LLMs) to process and answer queries. It focuses on integrating a relationship model and custom processing chains to handle various aspects of question-answering.

    Key functionalities include:
    - Initialization with paths and types for the relationship model.
    - Loading of the large language model (LLM) with specific configurations.
    - Execution of a custom processing chain, which integrates document processing and LLM querying.
    - Handling the process of asking questions to the LLM and retrieving answers.

    The class includes error handling and logging mechanisms to ensure robustness and traceability of operations. It's designed to handle complex querying processes by integrating multiple components for document processing and query answering in a cohesive workflow.
    """
class QASystem:
    def __init__(self, rel_model_path, rel_model_type):
        self.logger = setup_logger("Question-Answering_config", "Question-Answering")
        try:
                self.rel_model_path = rel_model_path
                self.rel_model_type = rel_model_type
                self.llm = self.load_llm()
        except Exception as e:
                self.logger.error(f"Invalid {self.__class__.__name__} configuration. Initialization failed: {e}")
                raise Exception(f"Initialization failed: {e}")
    
    def load_llm(self):
        try:
            llm = Llama(model_path=self.rel_model_path, f16_kv=True, n_ctx=1000)
            return llm
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Error loading LLM: {e}")
            raise Exception(f"Error loading LLM: {e}")

    def ask_question(self, prompt, llm, grammar=None):
        try:
            # output = self.custom_stuff_chain(reordered_docs=top_docs, llm_chain=llm_chain, query= prompt)
            output = llm.create_completion(prompt=prompt, grammar=grammar)
            return output
        
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Error asking questions to LLM: {e}")
            return {}


import asyncio
from pathlib import Path
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
from querent.processors.text_processor import TextProcessor
from querent.common.types.ingested_images import IngestedImages
import pytest
import uuid


@pytest.mark.asyncio
async def test_collect_and_ingest_pdf():
    # Set up the collector
    collector_factory = FSCollectorFactory()
    uri = Uri("file://" + str(Path("./tests/data/pdf/").resolve()))
    config = FSCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "root_path": uri.path,
            "name": "Local-config",
            "config": {},
            "uri": "file://",
        }
    )
    collector = collector_factory.resolve(uri, config)

    # Set up the ingestor
    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory(
        "pdf"
    )  # Notice the use of await here
    processor = TextProcessor()
    ingestor = await ingestor_factory.create("pdf", [processor])

    # Collect and ingest the PDF
    ingested_call = ingestor.ingest(collector.poll())
    counter = 0

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested is not None
            if ingested is not "" or ingested is not None:
                counter += 1
                if isinstance(ingested, IngestedImages):
                    print(ingested.page_num)
        assert (
            counter == 31
        )  # 30 pages in the PDF and 1 empty IngestedTokens to signify end of file

    await poll_and_print()  # Notice the use of await here


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_pdf())



import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import GithubConfig
from querent.common.uri import Uri
import uuid
from dotenv import load_dotenv

load_dotenv()


@pytest.fixture
def github_config():
    return GithubConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "github_username": os.getenv("USERNAME_GITHUB"),
            "repository": os.getenv("REPOSITORY_NAME_GITHUB"),
            "github_access_token": os.getenv("ACCESS_TOKEN_GITHUB"),
            "name": "Github-config",
            "config": {},
            "uri": "github://",
        }
    )


@pytest.mark.asyncio
async def test_github_collector(github_config):
    uri = Uri("github://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, github_config)
    assert collector is not None

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert not result.is_error()
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 6

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_github_collector())




import pytest
from querent.kg.ner_helperfunctions.filter_triples import TripleFilter


"""
    Asynchronous test function to validate the filtering and clustering of triples.

    This test initializes a TripleFilter with specified thresholds, filters the input triples,
    and then clusters the remaining triples. It asserts that the number of triples filtered
    and clustered matches the expected counts.

    Parameters:
    - triples (list of tuples): The input triples to be filtered and clustered.
    - score_threshold (float): The threshold for filtering based on entity scores.
    - attention_score_threshold (float): The threshold for filtering based on attention scores.
    - similarity_threshold (float): The threshold for filtering based on embedding similarity.
    - min_cluster_size (int): The minimum size of clusters to be formed.
    - min_samples (int): The number of samples in a neighborhood for a point to be considered a core point.
    - expected_filtered_count (int): The expected number of triples to be filtered out.
    - expected_clustered_count (int): The expected number of triples to be clustered.

    Assertions:
    - The number of triples filtered out matches the expected_filtered_count.
    - The number of triples after filtering matches the original count minus expected_filtered_count.
    - The number of triples clustered out matches the expected_clustered_count.
    - The number of triples after clustering matches the original count minus expected_clustered_count.
"""

# Define the test data and expected results
@pytest.mark.parametrize("triples, score_threshold, attention_score_threshold, similarity_threshold, min_cluster_size, min_samples, expected_filtered_count, expected_clustered_count", 
                        [([('eocene', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 1.0, "entity2_score": 0.69, "entity1_label": "B-GeoTime, B-GeoMeth", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "a Paleocene\\u2013Eocene Thermal Maximum (PETM) record", "entity2_nn_chunk": "a 543-m-thick (1780 ft) deep-marine section", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.2, "entity2_attnscore": 0.09, "pair_attnscore": 0.12, "entity1_embedding": [2.7497622966766357, -0.5454338788986206, 6.291350364685059, 8.717966079711914, 3.0895700454711914, 2.3901493549346924, 2.462153673171997, 2.4761993885040283, 3.185384511947632, 12.142127990722656], "entity2_embedding": [3.7323296070098877, -1.4545331001281738, 6.134968280792236, 8.18924331665039, 2.7829885482788086, 4.156617164611816, 1.9409900903701782, 2.868553876876831, 4.041325569152832, 13.211334228515625], "sentence_embedding": [8.273744583129883, 2.051804780960083, 9.59654426574707, -2.3267648220062256, 9.436607360839844, 10.587993621826172, -7.340267181396484, 4.733699321746826, 5.80666971206665, 5.723113536834717]}', 'ft'), 
           ('eocene', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 1.0, "entity2_score": 0.92, "entity1_label": "B-GeoTime, B-GeoMeth", "entity2_label": "B-GeoLoc", "entity1_nn_chunk": "a Paleocene\\u2013Eocene Thermal Maximum (PETM) record", "entity2_nn_chunk": "Mexico", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.2, "entity2_attnscore": 0.09, "pair_attnscore": 0.13, "entity1_embedding": [2.6488053798675537, -0.6798055768013, 6.543962478637695, 8.325488090515137, 3.091782331466675, 2.573601722717285, 2.505544900894165, 2.2582333087921143, 3.1716434955596924, 12.096002578735352], "entity2_embedding": [3.2900171279907227, -3.1185624599456787, 6.074228286743164, 8.012141227722168, 2.4355814456939697, 3.655385732650757, 2.7318809032440186, 2.2702815532684326, 3.1232528686523438, 12.755410194396973], "sentence_embedding": [8.274189949035645, 2.051741600036621, 9.596600532531738, -2.326831579208374, 9.588064193725586, 10.738533973693848, -7.178147792816162, 4.895790100097656, 5.51309061050415, 5.5418500900268555]}', 'mexico'), 
           ('eocene', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 1.0, "entity2_score": 0.98, "entity1_label": "B-GeoTime, B-GeoMeth", "entity2_label": "B-GeoMeth, B-GeoPetro", "entity1_nn_chunk": "a Paleocene\\u2013Eocene Thermal Maximum (PETM) record", "entity2_nn_chunk": "organic carbon stable isotopes", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.2, "entity2_attnscore": 0.21, "pair_attnscore": 0.21, "entity1_embedding": [2.792264461517334, -0.5597057938575745, 6.471883773803711, 8.395708084106445, 3.065075159072876, 2.7189583778381348, 2.3715662956237793, 2.3874168395996094, 3.2599542140960693, 12.358487129211426], "entity2_embedding": [3.6428165435791016, -0.1321650743484497, 5.9235968589782715, 8.318928718566895, 2.0116961002349854, 2.7097012996673584, 2.3251116275787354, 2.9655239582061768, 2.381112813949585, 11.753918647766113], "sentence_embedding": [8.273637771606445, 2.0518276691436768, 9.596525192260742, -2.3267481327056885, 9.410528182983398, 10.562018394470215, -7.365957736968994, 4.708027362823486, 5.797874927520752, 5.788640975952148]}', 'organic'), 
           ('ft', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 0.69, "entity2_score": 0.92, "entity1_label": "B-GeoMeth", "entity2_label": "B-GeoLoc", "entity1_nn_chunk": "a 543-m-thick (1780 ft) deep-marine section", "entity2_nn_chunk": "Mexico", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.09, "entity2_attnscore": 0.09, "pair_attnscore": 0.09, "entity1_embedding": [3.9145333766937256, -1.3819427490234375, 6.232261657714844, 7.93341064453125, 3.304962635040283, 3.6077494621276855, 1.7642797231674194, 2.5056090354919434, 4.156083106994629, 12.344873428344727], "entity2_embedding": [3.252547025680542, -2.6739959716796875, 6.196337699890137, 7.94928503036499, 2.37715482711792, 3.465878486633301, 2.636854887008667, 2.277503728866577, 2.9351625442504883, 13.037396430969238], "sentence_embedding": [8.273326873779297, 2.0518717765808105, 9.596479415893555, -2.326699733734131, 9.342904090881348, 10.494515419006348, -7.430538177490234, 4.64348030090332, 5.878793239593506, 5.812100410461426]}', 'mexico'), 
           ('ft', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 0.69, "entity2_score": 0.98, "entity1_label": "B-GeoMeth", "entity2_label": "B-GeoMeth, B-GeoPetro", "entity1_nn_chunk": "a 543-m-thick (1780 ft) deep-marine section", "entity2_nn_chunk": "organic carbon stable isotopes", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.09, "entity2_attnscore": 0.21, "pair_attnscore": 0.12, "entity1_embedding": [3.72121262550354, -1.2197364568710327, 6.335819244384766, 7.918046474456787, 3.3196980953216553, 4.441747665405273, 1.680811882019043, 2.783958673477173, 3.540374994277954, 12.884989738464355], "entity2_embedding": [3.5021800994873047, -0.06335429102182388, 5.944340229034424, 8.448829650878906, 1.9487491846084595, 2.593893527984619, 2.2817087173461914, 3.0285210609436035, 2.276712417602539, 11.722785949707031], "sentence_embedding": [8.273797035217285, 2.0517985820770264, 9.596549987792969, -2.3267695903778076, 9.452091217041016, 10.603425979614258, -7.32454776763916, 4.7494096755981445, 5.788027763366699, 5.748314380645752]}', 'organic'),
           ('mexico', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM.", "entity1_score": 0.92, "entity2_score": 0.98, "entity1_label": "B-GeoLoc", "entity2_label": "B-GeoMeth, B-GeoPetro", "entity1_nn_chunk": "Mexico", "entity2_nn_chunk": "organic carbon stable isotopes", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.09, "entity2_attnscore": 0.21, "pair_attnscore": 0.13, "entity1_embedding": [3.144767999649048, -2.7147321701049805, 6.127165794372559, 8.295948028564453, 2.8409361839294434, 3.421802282333374, 2.720162868499756, 1.9653069972991943, 3.038386106491089, 12.980123519897461], "entity2_embedding": [3.5817856788635254, -0.024259988218545914, 5.900996208190918, 8.518841743469238, 2.084223985671997, 2.620420455932617, 2.407585859298706, 3.089611768722534, 2.2546801567077637, 11.82196044921875], "sentence_embedding": [8.273309707641602, 2.051867723464966, 9.596478462219238, -2.3266983032226562, 9.337979316711426, 10.48964786529541, -7.436363697052002, 4.637650012969971, 5.869350433349609, 5.816451072692871]}', 'organic'), 
           ('tectonic', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "tectonic perturbations", "entity2_nn_chunk": "the upstream North American catchments", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.25, "entity2_attnscore": 0.11, "pair_attnscore": 0.15, "entity1_embedding": [3.6411795616149902, -1.5020869970321655, 5.696325778961182, 8.604700088500977, 1.5151718854904175, 3.677309274673462, 2.884289264678955, 2.873039484024048, 1.5940958261489868, 11.50014591217041], "entity2_embedding": [3.9356887340545654, -0.4274052679538727, 6.138661861419678, 8.16846752166748, 2.308915376663208, 4.314570426940918, 1.8623437881469727, 3.321876049041748, 3.1112611293792725, 12.501526832580566], "sentence_embedding": [8.273104667663574, 2.0518875122070312, 9.596617698669434, -2.3267197608947754, 7.928299903869629, 9.09572696685791, -9.138922691345215, 2.935373067855835, 4.520363807678223, 4.727715015411377]}', 'upstream'), 
           ('tectonic', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "tectonic perturbations", "entity2_nn_chunk": "the downstream sectors", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.25, "entity2_attnscore": 0.49, "pair_attnscore": 0.33, "entity1_embedding": [3.85038685798645, -1.3444548845291138, 5.504477024078369, 8.894646644592285, 1.6295241117477417, 3.37282395362854, 2.5772287845611572, 3.101915121078491, 1.5097862482070923, 11.506526947021484], "entity2_embedding": [3.9193172454833984, -1.6295044422149658, 5.632161617279053, 7.854856967926025, 1.1065350770950317, 2.7394042015075684, 2.5630269050598145, 2.9742021560668945, 1.9695501327514648, 11.684067726135254], "sentence_embedding": [8.274432182312012, 2.051668167114258, 9.596901893615723, -2.3269529342651367, 7.91616678237915, 9.085306167602539, -9.196913719177246, 2.8773529529571533, 4.365344524383545, 4.512395858764648]}', 'downstream'), 
           ('tectonic', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 0.98, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "tectonic perturbations", "entity2_nn_chunk": "the Gulf Coastal Plain", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.25, "entity2_attnscore": 0.24, "pair_attnscore": 0.24, "entity1_embedding": [3.5848329067230225, -1.487343668937683, 5.6952009201049805, 8.473453521728516, 1.5355619192123413, 3.6111767292022705, 2.654287099838257, 3.0325963497161865, 1.4424744844436646, 11.632696151733398], "entity2_embedding": [4.40934944152832, -2.6414687633514404, 5.719301700592041, 8.166884422302246, 1.9718804359436035, 3.195392608642578, 2.2786951065063477, 2.728088140487671, 3.600144386291504, 12.685139656066895], "sentence_embedding": [8.171407699584961, 2.0686159133911133, 9.576199531555176, -2.308645725250244, 8.752703666687012, 9.852017402648926, -7.976686954498291, 4.097899436950684, 2.8626537322998047, 3.544607162475586]}', 'coastal'), 
           ('upstream', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoMeth", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the upstream North American catchments", "entity2_nn_chunk": "the downstream sectors", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.11, "entity2_attnscore": 0.49, "pair_attnscore": 0.18, "entity1_embedding": [3.781593084335327, -0.8344725370407104, 6.194002151489258, 8.271034240722656, 2.546635389328003, 4.471875190734863, 1.8119113445281982, 3.2294750213623047, 2.757159471511841, 12.193735122680664], "entity2_embedding": [4.007772922515869, -1.539210557937622, 5.661635398864746, 7.687388896942139, 1.4495539665222168, 2.722914934158325, 2.4864203929901123, 2.8783230781555176, 2.188131809234619, 11.470794677734375], "sentence_embedding": [8.274091720581055, 2.0517280101776123, 9.596802711486816, -2.3268887996673584, 7.976433277130127, 9.145756721496582, -9.147445678710938, 2.926783561706543, 4.40964412689209, 4.408041000366211]}', 'downstream'), 
           ('upstream', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 0.98, "entity1_label": "B-GeoMeth", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the upstream North American catchments", "entity2_nn_chunk": "the Gulf Coastal Plain", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.11, "entity2_attnscore": 0.24, "pair_attnscore": 0.15, "entity1_embedding": [3.7742319107055664, -0.7761141657829285, 6.207308769226074, 8.10714340209961, 2.6537845134735107, 4.6391167640686035, 1.713283658027649, 3.2407264709472656, 3.028916597366333, 12.314061164855957], "entity2_embedding": [4.119009494781494, -2.451376438140869, 5.714200496673584, 8.16775131225586, 1.5697892904281616, 3.016974449157715, 2.3051607608795166, 2.6763205528259277, 4.043854236602783, 12.48615837097168], "sentence_embedding": [8.274133682250977, 2.0517232418060303, 9.596816062927246, -2.326892375946045, 7.954324722290039, 9.124204635620117, -9.18288803100586, 2.8913755416870117, 4.316779613494873, 4.332244873046875]}', 'coastal'), 
           ('downstream', '{"context": "In this study, we present evidence of a Paleocene\\u2013Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM) using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin.", "entity1_score": 1.0, "entity2_score": 0.98, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the downstream sectors", "entity2_nn_chunk": "the Gulf Coastal Plain", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.49, "entity2_attnscore": 0.24, "pair_attnscore": 0.32, "entity1_embedding": [3.9212284088134766, -1.6629095077514648, 5.521124362945557, 7.751332759857178, 1.503126621246338, 2.7465665340423584, 2.5462348461151123, 2.9435184001922607, 1.9544049501419067, 11.384193420410156], "entity2_embedding": [3.904635190963745, -2.3259449005126953, 5.728708267211914, 8.567526817321777, 1.8445405960083008, 3.1008949279785156, 2.3495540618896484, 2.6896140575408936, 4.384907245635986, 12.416000366210938], "sentence_embedding": [8.274145126342773, 2.0517215728759766, 9.59683609008789, -2.3269004821777344, 7.935454845428467, 9.104917526245117, -9.188153266906738, 2.8860979080200195, 4.3856353759765625, 4.518765926361084]}', 'coastal'), 
           ('basin', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the GoM basin", "entity2_nn_chunk": "deposition", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.26, "pair_attnscore": 0.26, "entity1_embedding": [3.779275417327881, -2.450312376022339, 5.564785957336426, 9.048267364501953, 1.158644199371338, 2.5670621395111084, 2.868556499481201, 2.8476035594940186, 2.626718044281006, 11.972245216369629], "entity2_embedding": [3.8690717220306396, -1.558268427848816, 5.745287895202637, 8.37027359008789, 1.8900668621063232, 2.226731777191162, 2.4036779403686523, 2.7072699069976807, 1.3761074542999268, 11.92121410369873], "sentence_embedding": [8.26048469543457, 2.0539541244506836, 9.59400463104248, -2.3245110511779785, 8.029595375061035, 9.17389965057373, -8.415410041809082, 3.659198522567749, 3.3177754878997803, 3.4890079498291016]}', 'deposition'), 
           ('basin', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the GoM basin", "entity2_nn_chunk": "a shale interval", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.1, "pair_attnscore": 0.14, "entity1_embedding": [4.119443416595459, -2.4170005321502686, 5.506854057312012, 8.978416442871094, 1.1511424779891968, 2.5173678398132324, 2.7195496559143066, 2.852993965148926, 2.2741880416870117, 12.252689361572266], "entity2_embedding": [3.9250998497009277, -0.9356150031089783, 5.388732433319092, 9.552205085754395, 1.5140143632888794, 2.5319619178771973, 2.372610330581665, 3.2776687145233154, 1.8205068111419678, 12.066429138183594], "sentence_embedding": [8.265246391296387, 2.0531725883483887, 9.594969749450684, -2.3253417015075684, 8.152375221252441, 9.298916816711426, -8.365521430969238, 3.7090201377868652, 3.5349483489990234, 3.0754904747009277]}', 'shale'), 
           ('basin', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "the GoM basin", "entity2_nn_chunk": "coarse-grained terrigenous material", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.05, "pair_attnscore": 0.09, "entity1_embedding": [4.108553409576416, -2.4838831424713135, 5.380271911621094, 9.477745056152344, 1.5617849826812744, 2.5357635021209717, 2.673309564590454, 2.7784769535064697, 2.1955718994140625, 12.205521583557129], "entity2_embedding": [4.0050835609436035, -1.1028891801834106, 5.440834045410156, 9.748600959777832, 1.7639741897583008, 2.5701112747192383, 2.4116194248199463, 2.9646735191345215, 2.2542827129364014, 11.504647254943848], "sentence_embedding": [8.258162498474121, 2.0543270111083984, 9.593584060668945, -2.324106454849243, 7.6216044425964355, 8.752443313598633, -8.413619995117188, 3.6613075733184814, 3.641577959060669, 3.4947078227996826]}', 'coarse'), 
           ('basin', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "the GoM basin", "entity2_nn_chunk": "upstream", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.09, "pair_attnscore": 0.13, "entity1_embedding": [3.736577033996582, -2.4790451526641846, 5.554500102996826, 9.023608207702637, 1.1513177156448364, 2.42797589302063, 2.8700180053710938, 2.8867266178131104, 2.2713053226470947, 12.193184852600098], "entity2_embedding": [3.824883222579956, -0.5760443806648254, 6.2958478927612305, 8.043895721435547, 2.6064090728759766, 4.462185382843018, 1.701485276222229, 3.265139102935791, 2.85941743850708, 12.199705123901367], "sentence_embedding": [8.258600234985352, 2.0542502403259277, 9.593637466430664, -2.3241846561431885, 7.877345085144043, 9.01498031616211, -8.370376586914062, 3.704374074935913, 3.5215251445770264, 3.322709083557129]}', 'upstream'), 
           ('deposition', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "deposition", "entity2_nn_chunk": "a shale interval", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.1, "pair_attnscore": 0.14, "entity1_embedding": [4.037097454071045, -1.5482368469238281, 5.687910079956055, 8.639769554138184, 2.0061299800872803, 2.580061912536621, 2.3091812133789062, 2.835251808166504, 0.9321686029434204, 11.964442253112793], "entity2_embedding": [3.823697328567505, -0.8554319143295288, 5.418757915496826, 9.303988456726074, 1.4667565822601318, 2.4567506313323975, 2.3202929496765137, 3.2809736728668213, 1.8063433170318604, 12.352104187011719], "sentence_embedding": [8.265887260437012, 2.0530619621276855, 9.595108985900879, -2.325455904006958, 8.23084545135498, 9.379379272460938, -8.350030899047852, 3.7244553565979004, 3.357015609741211, 3.0107994079589844]}', 'shale'), 
           ('deposition', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "deposition", "entity2_nn_chunk": "coarse-grained terrigenous material", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.05, "pair_attnscore": 0.09, "entity1_embedding": [4.022927761077881, -1.2513703107833862, 5.67182731628418, 8.537543296813965, 1.5244730710983276, 2.4617042541503906, 2.345921516418457, 3.041199207305908, 1.4364830255508423, 11.754050254821777], "entity2_embedding": [4.1893534660339355, -0.9747512936592102, 5.430134296417236, 9.750176429748535, 1.9102195501327515, 2.4549989700317383, 2.4048633575439453, 2.9038901329040527, 2.405259132385254, 11.941838264465332], "sentence_embedding": [8.260772705078125, 2.053903579711914, 9.594109535217285, -2.324568748474121, 7.807742595672607, 8.947274208068848, -8.485549926757812, 3.589221954345703, 3.6893534660339355, 3.214524030685425]}', 'coarse'), 
           ('deposition', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "deposition", "entity2_nn_chunk": "upstream", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.26, "entity2_attnscore": 0.09, "pair_attnscore": 0.13, "entity1_embedding": [4.127425193786621, -1.5212538242340088, 5.6802496910095215, 8.361967086791992, 1.9768118858337402, 2.3448410034179688, 2.362427234649658, 2.735774278640747, 1.5828312635421753, 11.465723037719727], "entity2_embedding": [4.109014511108398, -0.7797077894210815, 6.112752437591553, 8.394779205322266, 2.451874017715454, 4.320161819458008, 1.620858907699585, 3.317976236343384, 2.830693483352661, 12.318888664245605], "sentence_embedding": [8.26502513885498, 2.053211212158203, 9.594893455505371, -2.3252992630004883, 8.202630043029785, 9.348957061767578, -8.314787864685059, 3.7597386837005615, 3.485481023788452, 3.1397032737731934]}', 'upstream'), 
           ('shale', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "a shale interval", "entity2_nn_chunk": "coarse-grained terrigenous material", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.1, "entity2_attnscore": 0.05, "pair_attnscore": 0.07, "entity1_embedding": [3.8545055389404297, -1.1383901834487915, 5.378453731536865, 9.810062408447266, 1.592409610748291, 2.8200225830078125, 2.433525562286377, 3.1255710124969482, 1.8716100454330444, 11.913195610046387], "entity2_embedding": [4.333093643188477, -0.8805223703384399, 5.475193977355957, 9.671239852905273, 2.063199520111084, 2.3700921535491943, 2.3002822399139404, 2.8283228874206543, 2.4080417156219482, 11.918879508972168], "sentence_embedding": [8.267016410827637, 2.052884101867676, 9.595311164855957, -2.325650453567505, 8.268800735473633, 9.418998718261719, -8.360254287719727, 3.714193344116211, 3.7852654457092285, 2.7610220909118652]}', 'coarse'), 
           ('shale', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "a shale interval", "entity2_nn_chunk": "upstream", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.1, "entity2_attnscore": 0.09, "pair_attnscore": 0.09, "entity1_embedding": [3.913846731185913, -1.3515106439590454, 5.415563583374023, 9.736641883850098, 1.5567667484283447, 2.753096342086792, 2.371159076690674, 3.108031749725342, 1.6675533056259155, 12.004403114318848], "entity2_embedding": [3.9803621768951416, -0.6945693492889404, 6.077136993408203, 8.257105827331543, 2.2994720935821533, 4.3783793449401855, 1.7453073263168335, 3.397482395172119, 2.9937987327575684, 12.354580879211426], "sentence_embedding": [8.26601505279541, 2.053044557571411, 9.595142364501953, -2.32547926902771, 8.163567543029785, 9.313164710998535, -8.43984317779541, 3.634657859802246, 3.6852385997772217, 2.86068058013916]}', 'upstream'), 
           ('coarse', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoMeth", "entity1_nn_chunk": "coarse-grained terrigenous material", "entity2_nn_chunk": "upstream", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.05, "entity2_attnscore": 0.09, "pair_attnscore": 0.07, "entity1_embedding": [4.183037757873535, -0.95732182264328, 5.360871315002441, 10.020729064941406, 1.9280916452407837, 2.735372543334961, 2.475790023803711, 2.884106159210205, 2.3718998432159424, 11.750289916992188], "entity2_embedding": [4.050603866577148, -0.688696563243866, 6.089183807373047, 8.378361701965332, 2.33626127243042, 4.648577690124512, 1.70479416847229, 3.407353639602661, 3.1088290214538574, 12.333937644958496], "sentence_embedding": [8.26403522491455, 2.0533668994903564, 9.594727516174316, -2.3251283168792725, 8.122334480285645, 9.268819808959961, -8.390889167785645, 3.683662176132202, 3.695371150970459, 3.024846076965332]}', 'upstream'), 
           ('upstream', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 1.0, "entity1_label": "B-GeoMeth", "entity2_label": "B-GeoPetro", "entity1_nn_chunk": "upstream", "entity2_nn_chunk": "a particularly thick sedimentary section", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.09, "entity2_attnscore": 0.32, "pair_attnscore": 0.14, "entity1_embedding": [4.113377571105957, -0.4475870728492737, 6.095298767089844, 8.45187759399414, 2.4396491050720215, 4.143350601196289, 1.7025946378707886, 3.283315658569336, 2.8183791637420654, 12.168099403381348], "entity2_embedding": [4.4357590675354, -1.8888970613479614, 5.451931476593018, 9.404488563537598, 1.1754322052001953, 3.111313581466675, 2.1995394229888916, 3.109503746032715, 3.035334587097168, 12.325691223144531], "sentence_embedding": [8.266093254089355, 2.0530381202697754, 9.595128059387207, -2.3254902362823486, 8.218499183654785, 9.367105484008789, -8.362135887145996, 3.712357997894287, 3.6974105834960938, 2.9043288230895996]}', 'sedimentary'), 
           ('sedimentary', '{"context": "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom- modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi- ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental- scale paleo-drainage, which makes it of particular interest for paleoclimate and source- to-sink reconstructions.", "entity1_score": 1.0, "entity2_score": 0.87, "entity1_label": "B-GeoPetro", "entity2_label": "B-GeoPetro, I-GeoPetro", "entity1_nn_chunk": "a particularly thick sedimentary section", "entity2_nn_chunk": "the deep-sea fans", "file_path": "dummy_1_file.txt", "entity1_attnscore": 0.32, "entity2_attnscore": 0.12, "pair_attnscore": 0.17, "entity1_embedding": [4.744527816772461, -1.797021508216858, 5.398539066314697, 9.334464073181152, 1.169188141822815, 3.1069962978363037, 2.099773645401001, 3.097820281982422, 2.888684034347534, 12.149800300598145], "entity2_embedding": [3.2142386436462402, -2.473020553588867, 5.727625846862793, 9.082856178283691, 2.1118242740631104, 2.521404266357422, 2.7171003818511963, 2.5763702392578125, 3.8491625785827637, 12.245673179626465], "sentence_embedding": [8.262020111083984, 2.053694725036621, 9.594340324401855, -2.3247909545898438, 7.908104419708252, 9.051258087158203, -8.493917465209961, 3.5807642936706543, 3.4950454235076904, 3.304839611053467]}', 'sea')], 
            0.6, 0.1, 0.5, 5, 3, 2, 5), ] )
    
# Async test function for TripleFilter
@pytest.mark.asyncio
async def test_triple_filter(triples, score_threshold, attention_score_threshold, similarity_threshold, min_cluster_size, min_samples, expected_filtered_count, expected_clustered_count):
    # Initialize the TripleFilter with the provided thresholds
    triple_filter = TripleFilter(
        score_threshold=score_threshold,
        attention_score_threshold=attention_score_threshold,
        similarity_threshold=similarity_threshold,
        min_cluster_size=min_cluster_size,
        min_samples=min_samples,
        cluster_persistence_threshold=0.4
    )
    cluster_output = triple_filter.cluster_triples(triples)
    clustered_triples = cluster_output['filtered_triples']
    cluster_labels = cluster_output['cluster_labels']
    cluster_persistence = cluster_output['cluster_persistence']
    cluster_reduction_count = cluster_output['reduction_count']
    assert cluster_reduction_count == expected_clustered_count
    assert len(clustered_triples) == len(triples) - expected_clustered_count
    final_clustered_triples = triple_filter.filter_by_cluster_persistence(triples, cluster_persistence, cluster_labels)
    assert len(final_clustered_triples) == 5
    filtered_triples, reduction_count = triple_filter.filter_triples(final_clustered_triples)
    assert reduction_count == expected_filtered_count
    assert len(filtered_triples) == len(final_clustered_triples) - expected_filtered_count
    




import numpy as np
import pytest
from querent.kg.rel_helperfunctions.contextual_predicate import process_data
from querent.core.base_engine import BaseEngine
from typing import List, Tuple, Dict

"""
    Test the functionality of the ContextualPredicate class and its associated utility function.

    This test covers the following scenarios:
    1. Conversion of sample data tuples into a JSON format using the `convert_tuples_to_json` utility function.
    2. Verification of the correctness of the conversion against an expected JSON string.

    Assertions:
    - Ensure that the conversion of the sample data to JSON format is accurate and matches the expected JSON string.

"""
    
def test_contextual_predicate():
    sample_data = [[('temperatures', 'The Paleocene–Eocene Thermal Maximum (PETM) (ca. 56 Ma) was a rapid global warming event characterized by the rise of temperatures to5–9 °C (Kennett and Stott, 1991), which caused substantial environmental changes around the globe.', 'kenn', {'entity1_score': 0.91, 'entity2_score': 0.97, 'entity1_label': 'B-GeoMeth', 'entity2_label': 'B-GeoPetro', 'entity1_nn_chunk': 'temperatures', 'entity2_nn_chunk': 'Kennett', 'entity1_attnscore': 0.86, 'entity2_attnscore': 0.09, 'pair_attnscore': 0.16, 'entity1_embedding': [0.0975915715098381], 'entity2_embedding': [-1.2051959037780762]})]]

    result_list = process_data(sample_data, "dummy1.pdf")

    # Check if result_list is a list of tuples with three strings each
    if result_list and all(isinstance(item, tuple) and len(item) == 3 and all(isinstance(element, str) for element in item) for item in result_list):
        # If the condition is True, the assertion passes
        assert True
    else:
        # If the condition is False, the assertion fails
        assert False, "result_list is not a list of tuples like List[Tuple[str, str, str]]"




import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import DropboxConfig
from querent.common.uri import Uri
import uuid

from dotenv import load_dotenv

load_dotenv()


@pytest.fixture
def dropbox_config():
    return DropboxConfig(
        config_source={
            "backend": "dropbox",
            "id": str(uuid.uuid4()),
            "dropbox_app_key": os.getenv("DROPBOX_APP_KEY"),
            "dropbox_refresh_token": os.getenv("DROPBOX_REFRESH_TOKEN"),
            "dropbox_app_secret": os.getenv("DROPBOX_APP_SECRET"),
            "chunk_size": 1024,
            "folder_path": "",
            "name": "Dropbox-config",
            "config": {},
            "uri": "dropbox://",
        }
    )


@pytest.mark.asyncio
async def test_dropbox_collector(dropbox_config):
    uri = Uri("dropbox://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, dropbox_config)
    assert collector is not None

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert not result.is_error()
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 170

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_dropbox_collector())



from typing import AsyncGenerator

import dropbox
from dropbox.oauth import DropboxOAuth2FlowNoRedirect
from querent.common.types.collected_bytes import CollectedBytes
from querent.config.collector.collector_config import DropboxConfig
from querent.collectors.collector_base import Collector
from querent.config.collector.collector_config import CollectorBackend
from querent.collectors.collector_factory import CollectorFactory
from querent.common import common_errors
from querent.common.uri import Uri
from querent.logging.logger import setup_logger


class DropboxCollector(Collector):
    def __init__(self, config: DropboxConfig):
        self.dropbox_app_key = config.dropbox_app_key
        self.dropbox_app_secret = config.dropbox_app_secret
        self.folder_path = config.folder_path
        self.chunk_size = config.chunk_size
        self.refresh_token = config.dropbox_refresh_token
        self.logger = setup_logger(__name__, "DropboxCollector")
        self.dbx = None

    async def connect(self):
        try:
            if not self.dbx or not self.dbx.is_authenticated():
                # If the Dropbox SDK already has a valid access token, it won't refresh unnecessarily
                self.dbx = dropbox.Dropbox(
                    app_key=self.dropbox_app_key,
                    app_secret=self.dropbox_app_secret,
                    oauth2_refresh_token=self.refresh_token,
                )
        except dropbox.exceptions.AuthError as auth_error:
            self.logger.error(
                f"Authentication error during Dropbox connection: {auth_error}"
            )
            raise common_errors.ConnectionError(
                "Failed to authenticate with Dropbox."
            ) from auth_error
        except Exception as e:
            self.logger.error(f"Error connecting to Dropbox: {e}")
            raise common_errors.ConnectionError(
                "Failed to connect to Dropbox: {}".format(str(e))
            ) from e

    async def disconnect(self):
        try:
            if self.dbx:
                self.dbx.close()
        except Exception as e:
            self.logger.error(f"Error disconnecting from Dropbox: {e}")

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        try:
            await self.connect()
            files_list = self.dbx.files_list_folder(self.folder_path).entries
            for entry in files_list:
                name = f"/{entry.name}"
                try:
                    file_metadata, response = self.dbx.files_download(name)
                except dropbox.exceptions.AuthError as auth_error:
                    self.logger.warning(
                        f"Token expired. Refreshing access token and retrying."
                    )
                    self.connect()  # Attempt to refresh access token
                    file_metadata, response = self.dbx.files_download(name)

                file_content_bytes = response.content
                async for chunk in self.stream_blob(file_content_bytes):
                    yield CollectedBytes(file=entry.name, data=chunk)
                yield CollectedBytes(file=entry.name, data=None, eof=True)
        except dropbox.exceptions.ApiError as e:
            self.logger.error(f"Error polling Dropbox: {e}")
            raise common_errors.PollingError(
                "Failed to poll Dropbox: {}".format(str(e))
            ) from e
        finally:
            await self.disconnect()

    async def stream_blob(self, content_bytes):
        offset = 0
        while offset < len(content_bytes):
            yield content_bytes[offset : offset + self.chunk_size]
            offset = offset + self.chunk_size


class DropBoxCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.DropBox

    def resolve(self, uri: Uri, config: DropboxConfig) -> Collector:
        return DropboxCollector(config=config)



# import json
# from enum import Enum
# from typing import Union, Optional

# import requests
# from pydantic import BaseModel, Field

# import importlib
# from .pydantic_models_to_grammar import generate_gbnf_grammar_from_pydantic_models, generate_gbnf_grammar_and_documentation
# import concurrent.futures

# # Function to get completion on the llama.cpp server with grammar.
# def create_completion(prompt, grammar, n_predict):
#     headers = {"Content-Type": "application/json"}
#     data = {"prompt": prompt, "n_predict": n_predict, "temperature": 0}

#     response = requests.post("http://localhost:8080/completion", headers=headers, json=data)
#     data = response.json()
#     print(data)

#     print(data["content"])
#     return data["content"]

# def send_api_request(prompt):
#     headers = {"Content-Type": "application/json"}
#     data = {"prompt": prompt, "n_predict": 128, "temperature": 0}
    
#     response = requests.post("http://localhost:8080/completion", headers=headers, json=data)
#     response_data = response.json()
#     return response_data

# from pydantic import BaseModel, Field

# class Semantic_Triple(BaseModel):
#     """
#     Represents a semantic triple for controlling LLAMA output.
#     A semantic triple consists of a subject, subject type, object, object type,
#     predicate, and predicate type, allowing fine-grained control over language generation.
#     """
#     subject: str = Field(..., description="The subject of the semantic triple.")
#     subject_type: str = Field(..., description="The type of the subject.")
#     object: str = Field(..., description="The object of the semantic triple.")
#     object_type: str = Field(..., description="The type of the object.")
#     predicate: str = Field(..., description="The predicate of the semantic triple.")
#     predicate_type: str = Field(..., description="The type of the predicate.")



# gbnf_grammar, documentation = generate_gbnf_grammar_and_documentation([Semantic_Triple])

# # print("----------------------", gbnf_grammar)
# # print(documentation)
# # # JSON_Format: {{'subject': Subject, 'object': Object, 'predicate': Predicate, 'subject_type':Subject_Type, 'object_type':Object_Type, 'predicate_type':Predicate_Type  }}
# # # system_message = "You are an advanced AI, tasked to assist the user by calling functions in JSON format. The following are the available functions and their parameters and types:\n\n" + documentation
# prompt="""Please analyze the provided context and two entities. Only fill the values in place of output fields marked with {{output}} for the each respective key in the Json format described below and no need for any explanation.
# JSON_Format: {{'subject': {{output}}, 'object': {{output}}, 'predicate': {{output}}, 'subject_type':{{output}}, 'object_type':{{output}}, 'predicate_type':{{output}} }}
# Context: We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca.
# Entity 1: the GoM basin and Entity 2: coarse-grained terrigenous material
# Query: Determine which entity is the subject and which is the object in the context, along with the predicate, subject type, predicate type and object type.
# Answer:"""
# # # # user_message = "What is 42 * 42?"
# prompt = f"<|im_start|>user\n{prompt}<|im_end|>\n<|im_start|>assistant"

# text = create_completion(prompt=prompt, grammar=gbnf_grammar, n_predict=512)
# print(text)



# # Create two prompts and grammars for your API requests

# # prompt1 = """Please analyze the provided context and two entities. Just fill the output fields in the Json format described below and no need for explanation.
# #         JSON_Format: {{'subject': {{output}}, 'object': {{output}}, 'predicate': {{output}}, 'subject_type':{{output}}, 'object_type':{{output}}, 'predicate_type':{{output}} }}
# #         Context: We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca.
# #         Entity 1: deposition and Entity 2: a shale interval
# #         Query: Determine which entity is the subject and which is the object in the context, along with the predicate, subject type, predicate type and object type.
# #         Answer:"""


# # # Create a list of prompts and grammars
# # prompts = [prompt1, prompt]


# # # Create a ThreadPoolExecutor with 2 threads (for 2 simultaneous requests)
# # with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
# #     # Use executor.map to send the requests concurrently
# #     responses = list(executor.map(send_api_request, prompts))

# # # Process the responses
# # for i, response in enumerate(responses):
# #     print(f"Response {i+1}:")
# #     print(json.dumps(response, indent=4))


import asyncio
import pytest
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import JiraCollectorConfig
from querent.common.uri import Uri
import uuid
from dotenv import load_dotenv
import os

load_dotenv()


@pytest.fixture
def jira_config():
    return JiraCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "jira_server": "https://querent.atlassian.net/",
            "jira_username": "puneet@querent.xyz",
            "jira_api_token": os.getenv("JIRA_API_TOKEN"),
            "jira_project": "Querent1",
            "jira_query": "project=Querent1",
            "name": "Jira-config",
            "config": {},
            "uri": "jira://",
        }
    )


@pytest.mark.asyncio
async def test_jira_collector(jira_config):
    uri = Uri("jira://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, jira_config)
    assert collector is not None
    await collector.connect()

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert result is not None
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 1

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_jira_collector())



from uuid import uuid4
from datetime import datetime
import re

from querent.graph.errors import InvalidGraphNodeValueError, UnknownLiteralType


class Node(object):
    def __init__(self, value):
        if not isinstance(value, str):
            raise InvalidGraphNodeValueError()
        self._value = value

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        return self._value == other._value

    def __hash__(self):
        return hash(self.value)

    @property
    def value(self):
        return self._value

    def is_valid(self):
        raise NotImplementedError("Subclass should implement this.")


class URI(Node):
    def __init__(self, value):
        if isinstance(value, URI):
            super().__init__(value.value)
        else:
            super().__init__(value)

    def __eq__(self, other):
        if not isinstance(other, URI):
            return False
        return super().__eq__(other)

    def __hash__(self):
        return hash(self.value)

    def is_valid(self):
        return self.value is not None


class BNode(Node):
    def __init__(self, value=None):
        if isinstance(value, BNode):
            super().__init__(value.value)
        else:
            if not value:
                value = uuid4().hex
            super().__init__(value)

    def __eq__(self, other):
        if not isinstance(other, BNode):
            return False
        return super().__eq__(other)

    def __hash__(self):
        return hash(self.value)

    def is_valid(self):
        return self.value is not None


class Literal(Node):
    def __init__(self, value, lang=None, type_=None):
        if isinstance(value, Literal):
            super().__init__(value.value)
            self._lang = value.lang
            self._type = value._type
        else:
            super().__init__(value)
            self._lang = lang
            if type_ and isinstance(type_, str):
                type_ = LiteralType(type_)
            self._type = type_

    def __eq__(self, other):
        if not isinstance(other, Literal):
            return False
        return (
            super().__eq__(other)
            and self.lang == other.lang
            and self.type == other.type
        )

    def __hash__(self):
        return hash((self.value, self.lang, self.raw_type))

    def __str__(self):
        return self.value

    @property
    def lang(self):
        return self._lang

    @property
    def type(self):
        return self._type

    @property
    def raw_type(self):
        if self._type:
            return self._type.value

    def is_valid(self):
        if self.value is None:
            return False
        if self._type == LiteralType.string and not self._lang:
            return False
        return self._type.is_value_valid(self.value)


class __Type(type):
    def __getattr__(self, item):
        return LiteralType(item)


class LiteralType(URI, metaclass=__Type):
    valid_time_pattern = re.compile(
        r"[\-]?(\d{4})-((0[1-9])|(1[0-2]))-(0[1-9]|[12][0-9]|3[01])T(0[0-9]|1[0-9]|2[0-3])"
        r":(0[0-9]|1[0-9]|2[0-9]|3[0-9]|4[0-9]|5[0-9]):(0[0-9]|1[0-9]|2[0-9]|3[0-9]|4[0-9]|5[0-9])"
    )
    valid_month_pattern = re.compile(r"[\-]?(\d{4})-((0[1-9])|(1[0-2]))-(00)T00:00:00")
    valid_year_decade_millennium_pattern = re.compile(
        r"[\-]?(\d{4})-(00)-(00)T00:00:00"
    )
    valid_hundred_thousand_years_pattern = re.compile(
        r"[\-]?(\d{6,7})-(0[0|1])-(0[0|1])T00:00:00"
    )
    valid_million_billion_years = re.compile(
        r"[\-]?(\d{8}\d+)-(0[0|1])-(0[0|1])T00:00:00"
    )

    def __init__(self, s, common_check=True):
        self.common_check = common_check
        self._type = None
        super().__init__(self._resolve(s))

    def _resolve(self, s):
        if not isinstance(s, str):
            raise UnknownLiteralType()
        if not self.common_check:
            return s
        # complete xsd uri
        if s.startswith(self.xsd) and s[len(self.xsd) :] in self.xsd_tokens:
            self._type = s[len(self.xsd) :]
            return s
        # complete rdf uri
        elif s.startswith(self.rdf) and s[len(self.rdf)] in self.rdf_tokens:
            self._type = s[len(self.rdf)]
            return s
        # xsd prefix
        elif s.startswith("xsd:"):
            s = s[4:]
            if s in self.xsd_tokens:
                self._type = s
                return self.xsd + s
        # rdf prefix
        elif s.startswith("rdf:"):
            s = s[4:]
            if s in self.rdf_tokens:
                self._type = s
                return self.rdf + s
        # no prefix
        elif self._is_valid_field(s):
            if s in self.xsd_tokens:
                self._type = s
                return self.xsd + s
            elif s in self.rdf_tokens:
                self._type = s
                return self.rdf + s
        raise UnknownLiteralType()

    def _is_valid_field(self, s):
        return s in self.xsd_tokens or s in self.rdf_tokens

    def is_value_valid(self, s):
        """
        If value `s` if valid in current LiteralType
        """
        if self._type in self.value_validator:
            return self.value_validator[self._type](s)

        return True

    @staticmethod
    def _is_valid_date_time(s):
        if isinstance(s, datetime):
            return True

        try:
            # python 3.7
            datetime.fromisoformat(s)
            return True
        except Exception:
            pass

        valid_format = ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S.%f%z"]
        try:
            for f in valid_format:
                datetime.strptime(s, f)
                return True
        except Exception:
            pass

        validity_list = [
            LiteralType.valid_time_pattern.match(s),
            LiteralType.valid_month_pattern.match(s),
            LiteralType.valid_year_decade_millennium_pattern.match(s),
            LiteralType.valid_hundred_thousand_years_pattern.match(s),
            LiteralType.valid_million_billion_years.match(s),
        ]
        return any(validity_list)

    xsd = "http://www.w3.org/2001/XMLSchema#"
    rdf = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xsd_tokens = {
        "time",
        "date",
        "dateTime",
        "string",
        "normalizedString",
        "token",
        "language",
        "boolean",
        "decimal",
        "integer",
        "nonPositiveInteger",
        "long",
        "nonNegativeInteger",
        "negativeInteger",
        "int",
        "unsignedLong",
        "positiveInteger",
        "short",
        "unsignedInt",
        "byte",
        "unsignedShort",
        "unsignedByte",
        "float",
        "double",
        "base64Binary",
        "anyURI",
    }
    rdf_tokens = {"XMLLiteral", "HTML"}
    value_validator = {
        "dateTime": _is_valid_date_time.__func__,
    }
    # to_python_type = {
    #     'time': time,
    #     'date': date,
    #     'dateTime': datetime,
    #     'string': str,
    #     'normalizedString': str,
    #     'token': str,
    #     'language': str,
    #     'boolean': bool,
    #     'decimal': float,
    #     'integer': int,
    #     'nonPositiveInteger': int,
    #     'long': int,
    #     'nonNegativeInteger': int,
    #     'negativeInteger': int,
    #     'int': int,
    #     'unsignedLong': int,
    #     'positiveInteger': int,
    #     'short': int,
    #     'unsignedInt': int,
    #     'byte': int,
    #     'unsignedShort': int,
    #     'unsignedByte': int,
    #     'float': float,
    #     'double': float,
    #     'base64Binary': str,
    #     'anyURI': str,
    #     'XMLLiteral': Document,
    #     'HTML': DocumentFragment
    # }



import json

"""
    A class to convert triples into different JSON formats.

    This class provides static methods to convert triples (a tuple of subject, predicate, and object)
    into specific JSON structures suitable for graph and vector space representations.

    Methods:
    - _normalize_text(text, replace_space=False): Normalizes the text by converting it to lowercase
      and optionally replacing spaces with underscores.
    - _parse_json_str(json_str): Parses a JSON string and returns a Python dictionary.
    - convert_graphjson(triple): Converts a triple into a JSON object suitable for graph representation.
    - convert_vectorjson(triple): Converts a triple into a JSON object suitable for vector space representation.
    """

class TripleToJsonConverter:
    @staticmethod
    def _normalize_text(text, replace_space=False):
        normalized_text = text.lower()
        if replace_space:
            normalized_text = normalized_text.replace(" ", "_")
        return normalized_text

    @staticmethod
    def _parse_json_str(json_str):
        try:
            return json.loads(json_str)
        except json.JSONDecodeError as e:
            raise ValueError(f"Error decoding JSON: {e}")

    @staticmethod
    def convert_graphjson(triple):
        try:
            subject, json_str, object_ = triple
            predicate_info = TripleToJsonConverter._parse_json_str(json_str)
            if predicate_info is None:
                return {}

            json_object = {
                "subject": TripleToJsonConverter._normalize_text(subject),
                "subject_type": TripleToJsonConverter._normalize_text(predicate_info.get("subject_type", ""), replace_space=True),
                "object": TripleToJsonConverter._normalize_text(object_),
                "object_type": TripleToJsonConverter._normalize_text(predicate_info.get("object_type", ""), replace_space=True),
                "predicate": TripleToJsonConverter._normalize_text(predicate_info.get("predicate", "")),
                "predicate_type": TripleToJsonConverter._normalize_text(predicate_info.get("predicate_type", ""), replace_space=True),
                "sentence": predicate_info.get("context", "").lower()
            }

            return json_object
        except Exception as e:
            raise Exception(f"Error in convert_graphjson: {e}")

    @staticmethod
    def convert_vectorjson(triple):
        try:
            subject, json_str, object_ = triple
            data = TripleToJsonConverter._parse_json_str(json_str)
            if data is None:
                return {}

            id_format = f"{TripleToJsonConverter._normalize_text(subject)}_{TripleToJsonConverter._normalize_text(data.get('predicate', ''))}_{TripleToJsonConverter._normalize_text(object_)}"
            json_object = {
                "id": TripleToJsonConverter._normalize_text(id_format,replace_space=True),
                "embeddings": data.get("context_embeddings", []),
                "size": len(data.get("context_embeddings", [])),
                "namespace": TripleToJsonConverter._normalize_text(data.get("predicate", ""),replace_space=True)
            }

            return json_object
        except Exception as e:
            raise Exception(f"Error in convert_vectorjson: {e}")
        
        
        



class InvalidURIException(Exception):
    def __init__(self, uri):
        self.uri = uri
        super().__init__(self.uri)


class PrefixNotFoundException(Exception):
    def __init__(self, prefix):
        self.prefix = prefix
        super().__init__(self.prefix)


class InvalidGraphNodeValueError(Exception):
    def __init__(self, value):
        self.value = value
        super().__init__(self.value)


class UnknownLiteralType(Exception):
    def __init__(self):
        super().__init__()


class InvalidParameter(Exception):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(self.msg)


class SplitURIWithUnknownPrefix(Exception):
    def __init__(self):
        super().__init__()


class BadSchemaException(Exception):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(self.msg)



# UnknownMetricError(f"Unknown metric: {metric_name}")
class MetricError:
    UNKNOWNMETRICERROR = "Unknown Metric Error"
    UPDATEMETRICERROR = "Update Metric Error"


class UnknownMetricError(Exception):
    def __init__(self, message=None) -> None:
        super().__init__(message)
        self.error_code = MetricError.UNKNOWNMETRICERROR


class UpdateMetricError(Exception):
    def __init__(self, message=None) -> None:
        super().__init__(message)
        self.error_code = MetricError.UPDATEMETRICERROR


import os
from pydantic import BaseSettings
import yaml
from pydantic import BaseModel
from typing import Optional, List

from querent.common.types.config_keys import ConfigKey
from querent.config.workflow.workflow_config import WorkflowConfig
from querent.config.collector.collector_config import CollectorConfig
from querent.config.engine.engine_config import EngineConfig
from querent.config.resource.resource_config import ResourceConfig


class Config(BaseModel):
    version: float
    querent_id: str
    querent_name: str
    workflow: WorkflowConfig
    collectors: List[CollectorConfig]
    engines: List[EngineConfig]
    resource: Optional[ResourceConfig]

    def __init__(self, config_source=None, **kwargs):
        # if kwargs:
        #     raise ValueError(
        #         "Config values must be provided within a dictionary via 'config_source' parameter."
        #     )

        if config_source:
            config_data = self.load_config(config_source)
            super().__init__(**config_data)
            for config_key in ConfigKey:
                key = config_key.value
                if key in config_data:
                    setattr(self, key, config_data[key])

        # else:
        #     raise ValueError("Please pass config")

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    @classmethod
    def get_full_config(cls):
        return cls.config_data

    @classmethod
    def get(cls, key: ConfigKey, default=None):
        """
        Get a specific configuration value by key.
        Args:
            key (ConfigKey): The key for the configuration value.
            default: The default value to return if the key is not found.
        Returns:
            The configuration value if found, otherwise the default value.
        """
        return cls.config_data.get(key, default)

    @classmethod
    def has(cls, key: ConfigKey):
        """
        Check if a specific configuration key exists.
        Args:
            key (ConfigKey): The key to check.
        Returns:
            bool: True if the key exists, False otherwise.
        """
        return key in cls.config_data




# import json
# from llama_cpp import LlamaGrammar
# import re

# # grammar = LlamaGrammar.from_file(file = "./querent/kg/rel_helperfunctions/subject_object_grammar.gbnf")
# grammar = LlamaGrammar.from_file(file = "./querent/kg/rel_helperfunctions/json.gbnf")

# from llama_cpp import Llama
# # llm = Llama(model_path="./tests/llama-2-7b-chat.Q5_K_M.gguf")
# llm = Llama(model_path="/home/nishantg/Downloads/openhermes-2.5-mistral-7b.Q5_K_M.gguf")
# context = "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. Despite other thick PETM sections being observed elsewhere in the world, the one described in this study links with a continental-scale paleo-drainage, which makes it of particular interest for paleoclimate and source-to-sink reconstructions."
# # entity1 = "deposition"
# # entity2 = "a shale interval"
# entity1 = "the GoM basin"
# entity2 = "coarse-grained terrigenous material"
# # query = """Please analyze the provided context and two specified entities to construct a semantic triple. A semantic triple is a structure used in semantic analysis and consists of three parts: a subject, a predicate, and an object. The subject is the main entity being discussed, the predicate is the action or relationship that connects the subject and object, and the object is the entity that is affected by or related to the subject.
# # Context: {context}
# # Entity 1: {entity1} and Entity 2: {entity2}
# # Query: Using the semantic triple framework (Subject, Predicate, Object), determine which entity is the subject and which is the object in the context along with the predicate between the entities. Please also identify the subject type, predicate type and object type.
# # Answer:""".format(context = context, entity1=entity1, entity2=entity2)
# # output = llm.create_completion(
# #       prompt="""Please analyze the provided context and two entities
# #         Context: We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca.
# #         Entity 1: the GoM basin and Entity 2: coarse-grained terrigenous material
# #         Query: Determine which entity is the subject and which is the object in the context along with the predicate between the entities. Please also identify the subject type, predicate type and object type.
# #         Answer:""", grammar=grammar)

# # query = """Please analyze the provided context and two entities.
# # Context: {context}
# # Entity 1: {entity1} and Entity 2: {entity2}
# # Query: First, using the semantic triple (Subject, Predicate & Object) framework, determine which of the above entity is the subject and which entity is the object based on the above context. Then determine the predicate between these entities along with the subject type, object type & predicate type.
# # Answer:""".format(context = context, entity1=entity1, entity2=entity2)

# query = """
# Please analyze the provided context.
# Context: {context}
# Entity 1: {entity1} 
# Entity 2: {entity2}

# Query: First, using the semantic triple framework, determine which of the above entities is the subject and which is the object, and what is the predicate between these entities. The determine the subject type, object type, and predicate type.

# Answer:""".format(context=context, entity1=entity1, entity2=entity2)

# output = llm.create_completion(
#       prompt= query, grammar=grammar)
# print(output)
# choices_text = output['choices'][0]['text']
# extracted_data = choices_text
# print(extracted_data)



import os
import pytest
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
import shutil

@pytest.fixture(scope="function")
def embedding_store():
    my_directory_path = './tests/kg_tests/vectorstores/'
    store = EmbeddingStore(vector_store_path=my_directory_path)
    yield store

def test_create_and_save_index(embedding_store):
    sentences = """In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM) record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM)
using organic carbon stable isotopes and biostratigraphic constraints. We suggest that climate and tectonic perturbations in the upstream North American catchments can induce
a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom-
modation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi-
ment supply during the PETM, which is archived as a particularly thick sedimentary section in the deep-sea fans of the GoM basin. Despite other thick PETM sections being
observed elsewhere in the world, the one described in this study links with a continental-scale paleo-drainage, which makes it of particular interest for paleoclimate and source-
to-sink reconstructions."""
    
    embedding_store.create_index(sentences,verbose=True)
    index_base_name = 'my_FAISS_index'
    embedding_store.save_index(index_base_name)
    index_faiss_path = os.path.join(embedding_store.vector_store_path, "my_FAISS_index", "index.faiss")
    index_pkl_path = os.path.join(embedding_store.vector_store_path, "my_FAISS_index", "index.pkl")
    # Assert that both files exist
    assert os.path.isfile(index_faiss_path), "FAISS index file was not created."
    assert os.path.isfile(index_pkl_path), "Pickle index file was not created."
    # If the assertions pass, delete the my_FAISS_index folder
    index_directory_path = os.path.join(embedding_store.vector_store_path, "my_FAISS_index")
    if os.path.isdir(index_directory_path):
        shutil.rmtree(index_directory_path)


"""File to start workflow"""
import asyncio
import json
import uuid
import os
from querent.callback.event_callback_interface import EventCallbackInterface

from querent.config.config import Config
from querent.config.workflow.workflow_config import WorkflowConfig
from querent.config.collector.collector_config import CollectorConfig
from querent.config.core.llm_config import LLM_Config
from querent.config.core.gpt_llm_config import GPTConfig
from querent.collectors.collector_resolver import CollectorResolver
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
from querent.core.transformers.bert_llm import BERTLLM
from querent.core.transformers.gpt_llm import GPTLLM
from querent.common.types.querent_event import  EventState, EventType
from querent.querent.querent import Querent
from querent.querent.resource_manager import ResourceManager



async def start_workflow(config_dict: dict):
    #Start the workflow
    workflow_config = config_dict.get("workflow")
    workflow = WorkflowConfig(config_source=workflow_config)
    collector_configs = config_dict.get("collectors", [])
    collectors = []
    for collector_config in collector_configs: 
        collectors.append(CollectorConfig(config_source=collector_config).resolve())

    engine_configs = config_dict.get("engines", [])
    engines = []
    for engine_config in engine_configs:
        engine_config_source = engine_config.get("config",{})
        if engine_config["name"] == "knowledge_graph_using_openai_v1":
            engines.append(GPTConfig(config_source = engine_config_source))
        elif engine_config["name"] == "knowledge_graph_using_llama2_v1":
            engines.append(LLM_Config(config_source=engine_config_source))
    config_dict["engines"] = engines 
    config_dict["collectors"] = collectors 
    config_dict["workflow"] = workflow
    config = Config(config_source=config_dict)

    workflows = {"openai": start_gpt_workflow,
               "llama": start_llama_workflow}

    workflow = workflows.get(config.workflow.name)
    await workflow(config)


async def start_gpt_workflow(config: Config):
    collectors = []
    for collector_config in config.collectors:
        uri = Uri(collector_config.uri)
        collectors.append(CollectorResolver().resolve(uri=uri, config = collector_config))

    for collector in collectors:
        await collector.connect()

    result_queue = asyncio.Queue()

    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )

    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    resource_manager = ResourceManager()

    llm_instance = GPTLLM(result_queue, config.engines[0])

    llm_instance.subscribe(EventType.Graph, config.workflow.event_handler())
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )
    querent_task = asyncio.create_task(querent.start())
    token_feeder = asyncio.create_task(receive_token_feeder(resource_manager=resource_manager, config=config, result_queue=result_queue, state_queue=GPTLLM.state_queue))
    await asyncio.gather(ingest_task, querent_task, token_feeder)


#Config workflow channel for setting termination event
async def receive_token_feeder(resource_manager: ResourceManager, config: Config, result_queue: asyncio.Queue, state_queue: asyncio.Queue):
    while not resource_manager.querent_termination_event.is_set():
        tokens = config.workflow.tokens_feader.receive_tokens_in_python()
        message_state = config.workflow.channel.receive_in_python()
        if tokens is not None:
            await result_queue.put(tokens)
        if message_state is not None:
            await state_queue.put(message_state)

async def start_llama_workflow(config: Config):
    collectors = []
    for collector_config in config.collectors:
        uri = Uri(collector_config.uri)
        collectors.append(CollectorResolver().resolve(uri=uri, config = collector_config))

    for collector in collectors:
        await collector.connect()

    result_queue = asyncio.Queue()

    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )

    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    resource_manager = ResourceManager()

    llm_instance = BERTLLM(result_queue, config.engines[0])

    llm_instance.subscribe(EventType.Graph, config.workflow.event_handler())
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )
    querent_task = asyncio.create_task(querent.start())
    # token_feeder = asyncio.create_task(receive_token_feeder(resource_manager=resource_manager, config=config, result_queue=result_queue, state_queue=BERTLLM.state_queue))
    await asyncio.gather(ingest_task, querent_task) # Loop and do config.workflow.channel for termination event messageType = "stop"
    # , token_feeder)


# async def main():
#     class StateChangeCallback(EventCallbackInterface):
#         def handle_event(self, event_type: EventType, event_state: EventState):
#             assert event_state.event_type == EventType.Graph
#             triple = json.loads(event_state.payload)
#             print("triple: {}".format(triple))
#             assert isinstance(triple['subject'], str) and triple['subject']
#     config_source={
#         "version": 1.0,
#         "id": "ahdbfvd",
#         "querent_id": 12345,
#         "querent_name": "llama",
#         "workflow": {
#         "name": "llama",
#         "id": str(uuid.uuid4()),
#         "config": {},
#         "event_handler": StateChangeCallback
#     },
#         "collectors": [{
#             "id": str(uuid.uuid4()),
#             "name": "Local-config",
#             "config": {
#                 "root_path": "./tests/data/llm/pdf",
#             },
#             "backend":"localfile",
#         },
#         {
#             "id": str(uuid.uuid4()),
#             "name": "Drive-config",
#             "config": {
#                 "drive_refresh_token": os.getenv("DRIVE_REFRESH_TOKEN"),
#                 "drive_token": os.getenv("DRIVE_TOKEN"),
#                 "drive_scopes": os.getenv("DRIVE_SCOPES"),
#                 "chunk_size": 1024 * 1024,
#                 "drive_client_id": os.getenv("DRIVE_CLIENT_ID"),
#                 "drive_client_secret": os.getenv("DRIVE_CLIENT_SECRET"),
#                 "specific_file_type": "application/pdf",
#                 "folder_to_crawl": "1BtLKXcYBrS16CX0R4V1X7Y4XyO9Ct7f8",
#             },
#             "backend": "drive",
#         }],
        
#         "engines": [{ 
#         #https://github.com/Querent-ai/querent-rs/blob/main/src/config/config.rs#L172 con
#         "id": str(uuid.uuid4()),
#         "name": "knowledge_graph_using_llama2_v1",
#         "config": {"ner_model_name":"botryan96/GeoBERT",
#         "enable_filtering": True,
#         "filter_params": {
#                 'score_threshold': 0.5,
#                 'attention_score_threshold': 0.1,
#                 'similarity_threshold': 0.5,
#                 'min_cluster_size': 5,
#                 'min_samples': 3,
#                 'cluster_persistence_threshold':0.1
#         }
#     }}],        
#     }

#     await start_workflow(config_source)
    
# asyncio.run(main())



from typing import List, AsyncGenerator
from querent.common.types.collected_bytes import CollectedBytes
from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.processors.async_processor import AsyncProcessor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.processors.async_processor import AsyncProcessor
from querent.common.common_errors import (
    FileNotFoundError,
    IOError,
    UnidentifiedImageError,
)
import pytesseract
from PIL import Image, UnidentifiedImageError
import io
from querent.common.types.ingested_tokens import IngestedTokens
from querent.logging.logger import setup_logger


class ImageIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"jpg", "jpeg", "png"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return ImageIngestor(processors)


class ImageIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.JPG)
        self.processors = processors
        self.logger = setup_logger(__name__, "ImageIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        try:
            current_file = None

            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    continue

                if chunk_bytes.file != current_file:
                    if current_file:
                        text = await self.extract_and_process_image(
                            CollectedBytes(file=current_file, data=collected_bytes)
                        )
                        yield IngestedTokens(file=current_file, data=[text], error=None)
                        yield IngestedTokens(file=current_file, data=None, error=None)

                    current_file = chunk_bytes.file
                    collected_bytes = b""

                collected_bytes += chunk_bytes.data

            if current_file:
                text = await self.extract_and_process_image(
                    CollectedBytes(file=current_file, data=collected_bytes)
                )
                yield IngestedTokens(file=current_file, data=[text], error=None)
                yield IngestedTokens(file=current_file, data=None, error=None)

        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")

    async def extract_and_process_image(self, collected_bytes: CollectedBytes) -> str:
        text = await self.extract_text_from_image(collected_bytes)
        return await self.process_data(text)

    async def extract_text_from_image(self, collected_bytes: CollectedBytes) -> str:
        try:
            image = Image.open(io.BytesIO(collected_bytes.data))

        except FileNotFoundError as exc:
            raise FileNotFoundError(
                f"Given file with name as {collected_bytes.file} not found"
            ) from exc
        except IOError as exc:
            raise IOError(
                f"Unable to open given file with name as {collected_bytes.file}"
            ) from exc
        except UnidentifiedImageError as exc:
            raise UnidentifiedImageError(
                f"Unable to open the given file with given name {collected_bytes.file}"
            ) from exc

        text = pytesseract.image_to_string(image)
        return str(text).encode("utf-8").decode("unicode_escape")

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



import asyncio
from asyncio import Queue
import json
from pathlib import Path
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.querent_event import EventState, EventType
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.config.core.llm_config import LLM_Config
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid
from querent.common.types.file_buffer import FileBuffer
from querent.core.transformers.bert_llm import BERTLLM
from querent.querent.resource_manager import ResourceManager
from querent.querent.querent import Querent
import time

@pytest.mark.asyncio
async def test_ingest_all_async():
    # Set up the collectors
    directories = [ "./tests/data/llm/pdf/"]
    collectors = [
        FSCollectorFactory().resolve(
            Uri("file://" + str(Path(directory).resolve())),
            FSCollectorConfig(config_source={
                    "id": str(uuid.uuid4()),
                    "root_path": directory,
                    "name": "Local-config",
                    "config": {},
                    "backend": "localfile",
                    "uri": "file://",
                }),
        )
        for directory in directories
    ]

    # Set up the result queue
    result_queue = asyncio.Queue()

    # Create the IngestorFactoryManager
    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )
    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())
    resource_manager = ResourceManager()
    bert_llm_config = LLM_Config(
    ner_model_name="botryan96/GeoBERT",
    enable_filtering=True,
    filter_params={
            'score_threshold': 0.5,
            'attention_score_threshold': 0.1,
            'similarity_threshold': 0.5,
            'min_cluster_size': 5,
            'min_samples': 3,
            'cluster_persistence_threshold':0.2
        }
    )
    llm_instance = BERTLLM(result_queue, bert_llm_config)
    class StateChangeCallback(EventCallbackInterface):
        def handle_event(self, event_type: EventType, event_state: EventState):
            assert event_state.event_type == EventType.Graph
            triple = json.loads(event_state.payload)
            print("triple: {}".format(triple))
            assert isinstance(triple['subject'], str) and triple['subject']
    llm_instance.subscribe(EventType.Graph, StateChangeCallback())
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )
    querent_task = asyncio.create_task(querent.start())
    await asyncio.gather(ingest_task, querent_task)

if __name__ == "__main__":

    # Run the async function
    asyncio.run(test_ingest_all_async())



import asyncio
from typing import List

from querent.core.base_engine import BaseEngine
from querent.logging.logger import setup_logger
from querent.querent.resource_manager import ResourceManager


class AutoScaler:
    def __init__(
        self,
        resource_manager: ResourceManager,
        querenters: List[BaseEngine],
    ):
        self.resource_manager = resource_manager
        self.querenters = querenters
        self.logger = setup_logger(__name__, "auto_scaler")
        self.querent_termination_event = resource_manager.querent_termination_event
        self.worker_tasks: List[asyncio.Task] = []  # Store the worker tasks

    async def scale_querenters(self, total_requested_workers: int):
        try:
            current_total_workers = sum(
                querenter.num_workers for querenter in self.querenters
            )

            if total_requested_workers <= current_total_workers:
                # Scale up querenter workers
                self.worker_tasks = []
                for querenter in self.querenters:
                    if not querenter.termination_event.is_set():
                        workers = await querenter._start_workers()
                        # Create tasks for the workers and store them
                        worker_tasks = [
                            asyncio.create_task(worker) for worker in workers
                        ]
                        self.worker_tasks.extend(
                            worker_tasks
                        )  # Extend the list of worker tasks
                        self.logger.info(
                            f"Started {len(worker_tasks)} workers for {querenter.__class__.__name__}"
                        )

            else:
                raise Exception(
                    "Total requested workers exceed the current total workers."
                )

            # start and wait for the querenter workers
            if self.worker_tasks:
                await asyncio.wait(self.worker_tasks)
            else:
                self.querent_termination_event.set()
                self.logger.info("No workers to start")
        except Exception as e:
            self.querent_termination_event.set()
            self.logger.error(f"An error occurred during AutoScaler execution: {e}")
            raise e

    async def start(self):
        try:
            while (
                not self.querent_termination_event.is_set()
            ):  # Check termination_event
                # Calculate the total requested workers for all querenters
                total_requested_workers = sum(
                    querenter.num_workers for querenter in self.querenters
                )
                # Get the maximum allowed workers from the resource manager
                max_allowed_workers = (
                    await self.resource_manager.get_max_allowed_workers()
                )
                if total_requested_workers > max_allowed_workers:
                    self.logger.error(
                        "Total requested workers exceed the maximum allowed workers."
                    )
                    raise Exception(
                        "Total requested workers exceed the maximum allowed workers."
                    )

                # Scale querenter workers

                await self.scale_querenters(total_requested_workers)
                # Wait for a while before checking again (adjust this as needed)
                await asyncio.sleep(1)
        except Exception as e:
            self.logger.error(f"An error occurred during AutoScaler execution: {e}")
            raise e
        finally:
            # check if workers are still running
            if self.worker_tasks:
                for task in self.worker_tasks:
                    if not task.done():
                        task.cancel()
                        self.logger.info("Cancelled worker task")

    async def stop(self):
        self.logger.info("Stopping AutoScaler")
        # Stop the querenter workers
        await self.stop_querenters()
        self.logger.info("AutoScaler stopped")

    async def stop_querenters(self):
        for querenter in self.querenters:
            await querenter._stop_workers()



# import json
# from llama_cpp import LlamaGrammar
# import re

# # grammar = LlamaGrammar.from_file(file = "./querent/kg/rel_helperfunctions/subject_object_grammar.gbnf")
# grammar = LlamaGrammar.from_file(file = "./querent/kg/rel_helperfunctions/json.gbnf")

# from llama_cpp import Llama
# llm = Llama(model_path="./tests/llama-2-7b-chat.Q5_K_M.gguf", n_threads=3)
# context = "We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca."
# # entity1 = "deposition"
# # entity2 = "a shale interval"
# entity1 = "the GoM basin"
# entity2 = "coarse-grained terrigenous material"
# query = """Please analyze the provided context and two specified entities to construct a semantic triple. A semantic triple is a structure used in semantic analysis and consists of three parts: a subject, a predicate, and an object. The subject is the main entity being discussed, the predicate is the action or relationship that connects the subject and object, and the object is the entity that is affected by or related to the subject.
# Context: {context}
# Entity 1: {entity1} and Entity 2: {entity2}
# Query: Using the semantic triple framework (Subject, Predicate, Object), determine which entity is the subject and which is the object in the context along with the predicate between the entities. Please also identify the subject type, predicate type and object type.
# Answer:""".format(context = context, entity1=entity1, entity2=entity2)
# # output = llm.create_completion(
# #       prompt="""Please analyze the provided context and two entities
# #         Context: We suggest that climate and tectonic perturbations in the upstream North American catchments can induce a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accommodation and deposition of a shale interval when coarse-grained terrigenous material was trapped upstream at the onset of the PETM, and (2) a considerable increase in sediment supply during the PETM, which is archived as a particularly thick sedimentary section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca.
# #         Entity 1: the GoM basin and Entity 2: coarse-grained terrigenous material
# #         Query: Determine which entity is the subject and which is the object in the context along with the predicate between the entities. Please also identify the subject type, predicate type and object type.
# #         Answer:""", grammar=grammar)

# output = llm.create_completion(
#       prompt= query, grammar=grammar)
# print(output)
# choices_text = output['choices'][0]['text']
# extracted_data = choices_text
# print(extracted_data)



import asyncio
from pathlib import Path
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.common.types.ingested_tokens import IngestedTokens
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid


@pytest.mark.asyncio
async def test_collect_and_ingest_xlsx():
    collector_factory = FSCollectorFactory()
    uri = Uri("file://" + str(Path("./tests/data/xlsx/").resolve()))
    config = FSCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "root_path": uri.path,
            "name": "Local-config",
            "config": {},
            "uri": "file://",
        }
    )
    collector = collector_factory.resolve(uri, config)

    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("xlsx")
    ingestor = await ingestor_factory.create("xlsx", [])

    ingested_call = ingestor.ingest(collector.poll())
    counter = 0

    async def poll_and_print():
        nonlocal counter  # Use nonlocal to modify the outer counter variable
        async for ingested in ingested_call:
            assert ingested is not None
            assert isinstance(ingested, IngestedTokens)
            assert ingested.error is None
            assert ingested.file is not None
            counter += 1
        # 1 extra empty IngestedTokens signifying end of file
        assert counter == 2

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_xlsx())



# Function calling example using pydantic models.

import json
from enum import Enum
from typing import Union, Optional

import requests
from pydantic import BaseModel, Field

import importlib
from pydantic_models_to_grammar import generate_gbnf_grammar_and_documentation

# Function to get completion on the llama.cpp server with grammar.
def create_completion(prompt, grammar):
    headers = {"Content-Type": "application/json"}
    data = {"prompt": prompt, "grammar": grammar}

    response = requests.post("http://localhost:8080/completion", headers=headers, json=data)
    data = response.json()

    print(data["content"])
    return data["content"]


# A function for the agent to send a message to the user.
class SendMessageToUser(BaseModel):
    """
    Send a message to the User.
    """
    chain_of_thought: str = Field(..., description="Your chain of thought while sending the message.")
    message: str = Field(..., description="Message you want to send to the user.")

    def run(self):
        print(self.message)


# Enum for the calculator function.
class MathOperation(Enum):
    ADD = "add"
    SUBTRACT = "subtract"
    MULTIPLY = "multiply"
    DIVIDE = "divide"


# Very simple calculator tool for the agent.
class Calculator(BaseModel):
    """
    Perform a math operation on two numbers.
    """
    number_one: Union[int, float] = Field(..., description="First number.")
    operation: MathOperation = Field(..., description="Math operation to perform.")
    number_two: Union[int, float] = Field(..., description="Second number.")

    def run(self):
        if self.operation == MathOperation.ADD:
            return self.number_one + self.number_two
        elif self.operation == MathOperation.SUBTRACT:
            return self.number_one - self.number_two
        elif self.operation == MathOperation.MULTIPLY:
            return self.number_one * self.number_two
        elif self.operation == MathOperation.DIVIDE:
            return self.number_one / self.number_two
        else:
            raise ValueError("Unknown operation.")


# Here the grammar gets generated by passing the available function models to generate_gbnf_grammar_and_documentation function. This also generates a documentation usable by the LLM.
# pydantic_model_list is the list of pydanitc models
# outer_object_name is an optional name for an outer object around the actual model object. Like a "function" object with "function_parameters" which contains the actual model object. If None, no outer object will be generated
# outer_object_content is the name of outer object content.
# model_prefix is the optional prefix for models in the documentation. (Default="Output Model")
# fields_prefix is the prefix for the model fields in the documentation. (Default="Output Fields")
gbnf_grammar, documentation = generate_gbnf_grammar_and_documentation(
    pydantic_model_list=[SendMessageToUser, Calculator], outer_object_name="function",
    outer_object_content="function_parameters", model_prefix="Function", fields_prefix="Parameters")

print(gbnf_grammar)
print(documentation)

system_message = "You are an advanced AI, tasked to assist the user by calling functions in JSON format. The following are the available functions and their parameters and types:\n\n" + documentation

user_message = "What is 42 * 42?"
prompt = f"<|im_start|>system\n{system_message}<|im_end|>\n<|im_start|>user\n{user_message}<|im_end|>\n<|im_start|>assistant"

text = create_completion(prompt=prompt, grammar=gbnf_grammar)
# This should output something like this:
# {
#     "function": "calculator",
#     "function_parameters": {
#         "number_one": 42,
#         "operation": "multiply",
#         "number_two": 42
#     }
# }
function_dictionary = json.loads(text)
if function_dictionary["function"] == "calculator":
    function_parameters = {**function_dictionary["function_parameters"]}

    print(Calculator(**function_parameters).run())
    # This should output: 1764


# A example structured output based on pydantic models. The LLM will create an entry for a Book database out of an unstructured text.
class Category(Enum):
    """
    The category of the book.
    """
    Fiction = "Fiction"
    NonFiction = "Non-Fiction"


class Book(BaseModel):
    """
    Represents an entry about a book.
    """
    title: str = Field(..., description="Title of the book.")
    author: str = Field(..., description="Author of the book.")
    published_year: Optional[int] = Field(..., description="Publishing year of the book.")
    keywords: list[str] = Field(..., description="A list of keywords.")
    category: Category = Field(..., description="Category of the book.")
    summary: str = Field(..., description="Summary of the book.")


# We need no additional parameters other than our list of pydantic models.
gbnf_grammar, documentation = generate_gbnf_grammar_and_documentation([Book])

system_message = "You are an advanced AI, tasked to create a dataset entry in JSON for a Book. The following is the expected output model:\n\n" + documentation

text = """The Feynman Lectures on Physics is a physics textbook based on some lectures by Richard Feynman, a Nobel laureate who has sometimes been called "The Great Explainer". The lectures were presented before undergraduate students at the California Institute of Technology (Caltech), during 1961–1963. The book's co-authors are Feynman, Robert B. Leighton, and Matthew Sands."""
prompt = f"<|im_start|>system\n{system_message}<|im_end|>\n<|im_start|>user\n{text}<|im_end|>\n<|im_start|>assistant"

text = create_completion(prompt=prompt, grammar=gbnf_grammar)

json_data = json.loads(text)

print(Book(**json_data))



import re
from typing import List
from nltk.corpus import wordnet as wn
"""
    A class designed to extract sentences from text that contain specified fixed relationships,
    taking into account synonyms for more comprehensive matching.

    This class utilizes regular expressions, spaCy's NLP capabilities, and the WordNet 
    database to identify and extract sentences from a given text which include any of the 
    user-specified relationships or their synonyms. It's particularly useful for focusing on
    specific types of relationships within large volumes of text.

    Attributes:
        nlp (spacy.Language): An instance of spaCy's language model.
        fixed_relationships (List[str]): A list of relationships (as strings) to search for in the text.
        relationship_patterns (List[re.Pattern]): Compiled regex patterns for the fixed relationships and their synonyms, 
                                                  enabling case-insensitive searching.

    Methods:
        create_patterns_with_synonyms(relationships: List[str]) -> List[re.Pattern]:
            Generates and compiles regex patterns that include synonyms of the specified relationships.

        find_relationship_sentences(text: str) -> str:
            Identifies and returns sentences from the provided text that contain any of the 
            fixed relationships or their synonyms.

        measure_reduction(original_text: str, reduced_text: str) -> float:
            Calculates the percentage reduction in text length after extracting relevant sentences.
"""

class FixedRelationshipExtractor:
    def __init__(self, fixed_relationships: List[str], model=None):
        self.nlp = model
        self.fixed_relationships = fixed_relationships
        self.relationship_pattern = self.create_combined_pattern_with_synonyms(fixed_relationships)

    def create_combined_pattern_with_synonyms(self, relationships):
        all_synonyms = set()
        for relationship in relationships:
            all_synonyms.add(relationship)
            for syn in wn.synsets(relationship, pos=wn.VERB):
                for l in syn.lemmas():
                    all_synonyms.add(l.name().replace('_', ' '))
        combined_pattern = '|'.join(map(re.escape, all_synonyms))
        return re.compile(r'\b(?:' + combined_pattern + r')\b', re.IGNORECASE)

    def find_relationship_sentences(self, text: str, chunk_size=1000) -> str:
        doc = self.nlp(text)
        relevant_sentences = set()
        prev_sentence = None

        for i in range(0, len(doc), chunk_size):
            chunk = doc[i:i+chunk_size]
            sentences = list(chunk.sents)

            for j, sentence in enumerate(sentences):
                sentence_text = sentence.text
                if self.relationship_pattern.search(sentence_text):
                    # Add the previous, current, and next sentences
                    if prev_sentence:
                        relevant_sentences.add(prev_sentence.text)
                    relevant_sentences.add(sentence_text)
                    if j < len(sentences) - 1:  # Check if there is a next sentence
                        relevant_sentences.add(sentences[j + 1].text)

                prev_sentence = sentence

        return ' '.join(sorted(relevant_sentences))


    def measure_reduction(self, original_text: str, reduced_text: str) -> float:
        original_length = len(original_text)
        reduced_length = len(reduced_text)
        reduction_percentage = ((original_length - reduced_length) / original_length) * 100
        return reduction_percentage



from abc import ABC, abstractmethod
import asyncio
from querent.callback.event_callback_dispatcher import EventCallbackDispatcher
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.common.types.ingested_images import IngestedImages
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_event import EventState, EventType
from querent.common.types.querent_queue import QuerentQueue
from querent.config.engine.engine_config import EngineConfig
from querent.logging.logger import setup_logger
from querent.common.types.ingested_images import IngestedImages
from querent.common.types.ingested_images import IngestedImages
import uuid

"""
    BaseEngine is an abstract base class that provides the foundational structure and methods 
    for processing tokens asynchronously and managing event states in a queue-based system.

    Attributes:
        input_queue (QuerentQueue): The queue containing input data to be processed.
        config (EngineConfig): The configuration for the engine.

    Methods:
        process_tokens(data: IngestedTokens) -> EventState:
            Abstract method to process tokens asynchronously.

        process_messages(data: IngestedMessages) -> EventState:
            Abstract method to process chat asynchronously.

        validate() -> bool:
            Abstract method to validate the configuration of the engine.

        set_state(new_state: EventState) -> None:
            Set the state to a new value.

        _listen_for_state_changes() -> None:
            Listen for changes in the state and notify subscribers.

        _worker() -> None:
            Worker task to process tokens and manage retries.

        _start_workers(number_of_workers: int) -> list:
            Start the specified number of worker tasks.

        _stop_workers() -> None:
            Stop all worker tasks.

        subscribe(event_type: EventType, callback: Callable) -> None:
            Subscribe to a specific event type.

        _notify_subscribers(event_type: EventType, event_state: EventState) -> None:
            Notify subscribers when an event occurs.

        set_termination_event() -> None:
            Set the termination event to signal termination of workers and listeners.
    """


class BaseEngine(ABC):
    def __init__(
        self,
        input_queue: QuerentQueue,
        config: EngineConfig = EngineConfig(
            config_source={
                "id": str(uuid.uuid4()),
                "name": "BaseEngine",
                "description": "Base Engine",
                "version": "0.0.1",
            }
        ),
        **kwargs,
    ):
        super().__init__(**kwargs)  # Call the super constructor first
        self.input_queue = input_queue
        self.termination_event = asyncio.Event()
        self.state_queue = QuerentQueue()
        self.num_workers = config.num_workers
        self.max_retries = config.max_retries
        self.retry_interval = config.retry_interval
        self.message_throttle_limit = config.message_throttle_limit
        self.message_throttle_delay = config.message_throttle_delay
        self.logger = setup_logger(config.logger, f"{__name__}.base_engine")
        self.callback_dispatcher = EventCallbackDispatcher()

    @abstractmethod
    async def process_tokens(self, data: IngestedTokens):
        """
        Process tokens asynchronously.
        Args:
            data (IngestedTokens): The input data to process.
        Returns:
            EventState: The state of the event is set with the event type and the timestamp
            of the event and set using `self.set_state(event_state)`.
        """
        raise NotImplementedError

    @abstractmethod
    async def process_messages(self, data: IngestedMessages):
        """
        Process chat asynchronously.
        Args:
            data (IngestedTokens): The input data to process.
        Returns:
            EventState: The state of the event is set with the event type and the timestamp
            of the event and set using `self.set_state(event_state)`.
        """
        raise NotImplementedError

    @abstractmethod
    async def process_code(self, data: IngestedCode):
        """
        Process coding files asynchronously.
        Args:
            data (IngestedCode): The input data to process.
        Returns:
            EventState: The state of the event is set with the event type and the timestamp
            of the event and set using `self.set_state(event_state)`.
        """
        raise NotImplementedError

    @abstractmethod
    async def process_images(self, data: IngestedImages):
        """
        Process image files asynchronously.
        Args:
            data (IngestedImage): The input data to process.
        Returns:
            EventState: The state of the event is set with the event type and the timestamp
            of the event and set using `self.set_state(event_state)`.
        """
        pass

    @abstractmethod
    def validate(self) -> bool:
        """
        Validate the LLM.
        Returns:
            bool: True if the LLM is valid, False otherwise.
        """
        raise NotImplementedError

    def set_termination_event(self):
        """
        Set termination event: Extensions of base engines must call this method to signal termination of workers and listeners.
        """
        self.termination_event.set()

    def subscribe(self, event_type: EventType, callback: EventCallbackInterface):
        """
        Subscribe to a specific event type.
        Args:
            event_type (EventType): The type of event to subscribe to (e.g., "token_processed").
            callback (Callable): The callback function to be invoked when the event occurs.
        """
        self.callback_dispatcher.register_callback(event_type, callback)

    async def set_state(self, new_state: EventState):
        """
        Set the state to a new value.
        Args:
            new_state (EventState): The new state.
        """
        if isinstance(new_state, EventState):
            await self.state_queue.put(new_state)
        else:
            raise Exception(
                f"Bad state type {type(new_state)} for {self.__class__.__name__}. Supported type: {EventState}"
            )

    """
        The following methods enables independent processing of tokens and messages.
    """

    async def _listen_for_state_changes(self):
        while not self.state_queue.empty() or not self.termination_event.is_set():
            new_state = await self.state_queue.get()
            if isinstance(new_state, EventState):
                if new_state.payload == "Terminate":
                    break
                await self._notify_subscribers(new_state.event_type, new_state)
            else:
                raise Exception(
                    f"Bad state type {type(new_state)} for {self.__class__.__name__}. Supported type: {EventState}"
                )
            await self.state_queue.task_done()

    async def _notify_subscribers(self, event_type: EventType, event_state: EventState):
        """
        Notify subscribers when an event occurs.
        Args:
            event_type (EventType): The type of event that occurred.
            event_data (EventState): The data associated with the event.
        """
        await self.callback_dispatcher.dispatch_event(event_type, event_state)
        await self.callback_dispatcher.dispatch_webhook(event_type, event_state)

    async def _worker(self):
        try:
            if not self.validate():
                self.logger.error(
                    f"Invalid {self.__class__.__name__} configuration. Please check the configuration."
                )
                raise ValueError(
                    f"Invalid {self.__class__.__name__} configuration. Please check the configuration."
                )

            state_listener = asyncio.create_task(self._listen_for_state_changes())

            async def _inner_worker():
                current_message_total = 0
                while not self.termination_event.is_set():
                    retries = 0
                    data = await self.input_queue.get()
                    try:
                        if isinstance(data, IngestedMessages):
                            await self.process_messages(data)
                        elif isinstance(data, IngestedTokens):
                            await self.process_tokens(data)
                        elif isinstance(data, IngestedImages):
                            await self.process_images(data)
                        elif isinstance(data, IngestedCode):
                            await self.process_code(data)
                        elif data is None:
                            self.termination_event.set()
                            current_state = EventState(EventType.Terminate,1.0, "Terminate", "temp.txt")
                            await self.set_state(new_state=current_state)

                        else:
                            raise Exception(
                                f"Invalid data type {type(data)} for {self.__class__.__name__}. Supported type: {IngestedTokens, IngestedMessages}"
                            )
                    except Exception as e:
                        self.logger.error(
                            f"Error processing tokens: {e}. Retrying ({retries}/{self.max_retries})"
                        )
                        retries += 1

                        if retries > self.max_retries:
                            self.logger.error(
                                f"Error processing tokens: {e}. Max retries reached. Terminating."
                            )
                            break
                        
                        await asyncio.sleep(self.retry_interval)
                    self.input_queue.task_done()
                    current_message_total += 1

                    if current_message_total >= self.message_throttle_limit:
                        await asyncio.sleep(self.message_throttle_delay)
                        current_message_total = 0

            await asyncio.gather(state_listener, _inner_worker())
        except Exception as e:
            self.logger.error(f"Error while processing tokens: {e}")
        finally:
            self.logger.info(f"Stopping worker for {self.__class__.__name__}")
            self.logger.info(f"Stopped worker for {self.__class__.__name__}")
            self.termination_event.set()

    async def _start_workers(self):
        self.workers = [self._worker() for _ in range(self.num_workers)]
        return self.workers

    async def _stop_workers(self):
        try:
            self.termination_event.set()
        except Exception as e:
            self.logger.error(f"Error while stopping workers: {e}")



import asyncio
import fnmatch
import re
from pathlib import Path
from typing import AsyncGenerator
from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.uri import Uri
from querent.config.collector.collector_config import CollectorBackend, FSCollectorConfig
import aiofiles
from querent.common import common_errors


class FSCollector(Collector):
    def __init__(self, config: FSCollectorConfig):
        self.root_dir = Path(config.root_path)
        self.items_to_ignore = []
        self.chunk_size = config.chunk_size
        try:
            with open("./.gitignore", "r", encoding="utf-8") as gitignore_file:
                self.items_to_ignore = set(
                    gitignore_file.read().replace("/", "").replace("*", "").splitlines()
                )
        except Exception as e:
            self.items_to_ignore = []

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    # collect those files

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        async for file_path in self.walk_files(self.root_dir):
            try:
                async with aiofiles.open(file_path, "rb") as file:
                    file_path_str = str(file_path)
                    async for chunk in self.read_chunks(file):
                        yield CollectedBytes(file=file_path_str, data=chunk, error=None)
                    yield CollectedBytes(
                        file=file_path_str, data=None, error=None, eof=True
                    )
            except PermissionError as exc:
                raise common_errors.PermissionError(
                    f"Unable to open this file {file_path}, getting error as {exc}"
                ) from exc
            except OSError as exc:
                raise common_errors.OSError(
                    f"Getting OS Error on file {file_path}, as {exc}"
                ) from exc
            except Exception as exc:
                raise common_errors.UnknownError(
                    f"Getting unknown error on file {file_path}, as {exc}"
                ) from exc

    async def read_chunks(self, file):
        while True:
            chunk = await file.read(self.chunk_size)
            if not chunk:
                break
            yield chunk
        await file.close()

    async def walk_files(self, root: Path) -> AsyncGenerator[Path, None]:
        for item in root.iterdir():
            item_split = set(str(item).split("/"))
            item_split.discard("")  # Use discard instead of remove to avoid KeyError
            if item_split and item_split.intersection(self.items_to_ignore):
                continue
            if item.is_file():
                yield item
            elif item.is_dir():
                async for file_path in self.walk_files(item):
                    yield file_path


class FSCollectorFactory(CollectorFactory):
    def __init__(self):
        pass

    def backend(self) -> CollectorBackend:
        return CollectorBackend.LocalFile

    def resolve(self, uri: Uri, config: FSCollectorConfig) -> Collector:
        return FSCollector(config)



# Description: Querent is an async data dynamo that can be used to collect, ingest, process, and store data.
#              It is designed to be used in a modular fashion, allowing for easy extensibility and customization.
#              It is also designed for data intensive applications, and is built on top of asyncio
#              to allow for high throughput and scalability.

# Version: 0.1.0



from querent.config.core.llm_config import LLM_Config

class GPTConfig(LLM_Config):
    name: str = "OPENAIEngine"
    description: str = "An engine for NER using BERT and knowledge graph operations using OPENAI"
    version: str = "0.0.1"
    logger: str = "OPENAI.engine_config"
    ner_model_name: str = "botryan96/GeoBERT"
    rel_model_name: str = "gpt-3.5-turbo"
    requests_per_minute: int = 3
    openai_apikey: str = ""

    def __init__(self, config_source=None, **kwargs):
        config_data = {}
        config_data.update(kwargs)
        if config_source:
            config_data = self.load_config(config_source)
        if "config" in config_data:
            config_data.update(config_data["config"])
        super().__init__(**config_data)


from typing import List, AsyncGenerator
import io
import pandas as pd

from querent.ingestors.ingestor_factory import IngestorFactory
from querent.ingestors.base_ingestor import BaseIngestor
from querent.processors.async_processor import AsyncProcessor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.types.ingested_tokens import (
    IngestedTokens,
)  # Added import for the return type
from querent.logging.logger import setup_logger


class XlsxIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"xlsx"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return XlsxIngestor(processors)


class XlsxIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.XLSX)
        self.processors = processors
        self.logger = setup_logger(__name__, "XlsxIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        current_file = None
        collected_bytes = b""
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    continue
                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    async for data in self.extract_and_process_xlsx(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield IngestedTokens(
                            file=current_file,
                            data=[data],
                            error=None,
                        )
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                collected_bytes += chunk_bytes.data
        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            async for data in self.extract_and_process_xlsx(
                CollectedBytes(file=current_file, data=collected_bytes)
            ):
                yield IngestedTokens(
                    file=current_file,
                    data=[data],
                    error=None,
                )
            yield IngestedTokens(file=current_file, data=None, error=None)

    async def extract_and_process_xlsx(
        self, collected_bytes: CollectedBytes
    ) -> AsyncGenerator[pd.DataFrame, None]:
        df = await self.extract_text_from_xlsx(collected_bytes)
        processed_text = await self.process_data(df.to_string())
        yield processed_text

    async def extract_text_from_xlsx(
        self, collected_bytes: CollectedBytes
    ) -> pd.DataFrame:
        excel_buffer = io.BytesIO(collected_bytes.data)
        dataframe = pd.read_excel(excel_buffer)
        return dataframe

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")
